Require Import List.

Require Import shared_defs.
Require Import shared_tactics.
Require Import step_defs.
Require Import teqt_defs.
Require Import step_theory.
Require Import teqt_db_basics.

Fixpoint translate_exp (e : exp) : texp :=
 match e with
   | exp_var x         => texp_var x
   | exp_fun a         => texp_fun (translate_exp a)
   | exp_lam a         => texp_lam (translate_exp a)
   | exp_nrec a        => texp_nrec (translate_exp a)
   | exp_app a b       => texp_app (translate_exp a) (translate_exp b)
   | exp_zero          => texp_zero
   | exp_succ a        => texp_succ (translate_exp a)
   | exp_ncase a b1 b2 => texp_ncase (translate_exp a) (translate_exp b1) (translate_exp b2)
   | exp_pair a b      => texp_pair (translate_exp a) (translate_exp b)
   | exp_pcase a b     => texp_pcase (translate_exp a) (translate_exp b)
   | exp_join          => texp_join
(*
   | exp_terminates    => texp_terminates
*)
   | _                 => texp_zero (* dummy value *)
  end.

Fixpoint translate_typ (e : exp) (th : theta) : ttyp :=
  match e with
   | exp_arr A B   => ttyp_arr th (translate_typ A th_log) (translate_typ B th)
   | exp_at A th'  => translate_typ A th'
   | exp_nat       => ttyp_nat
   | exp_sigma A B => ttyp_sigma (translate_typ A th_log) (translate_typ B th)
   | exp_eq a b    => ttyp_eq (translate_exp a) (translate_exp b)
(*
   | exp_Terminates a A => ttyp_terminates (translate_exp a) (translate_typ A th_log)
*)
   | _             => ttyp_nat (* dummy value *)
  end.

Fixpoint translate_ctx (G : ctx) : tctx :=
  match G with
   | nil => nil
   | (th,A)::G' => translate_typ A th :: translate_ctx G'
  end.

Lemma translate_ctx_app : forall G G',
  translate_ctx (G++G') = translate_ctx G ++ translate_ctx G'.
Proof.
  induction G as [ | [th A] G IH ];
     simpl; intros; rewrite_ih; auto.
Qed.

Lemma translate_ctx_length : forall G,
  length (translate_ctx G) = length G.
Proof.
  induction G  as [ | [th A] G IH ];   
    simpl; auto with arith.
Qed.
 
Lemma translate_FO : forall A th th', 
  FO A -> translate_typ A th = translate_typ A th'.
Proof.
 induction 1; simpl; auto; congruence.
Qed.

Lemma translate_val : forall v,
  val v -> tval (translate_exp v).
Proof.
  induction 1; simpl; auto.
Qed.

Lemma translate_typ_of_not_typelike :
  forall a th, ~typelike a -> translate_typ a th = ttyp_nat.
Proof.
induction a; simpl in *; intros; auto; tauto.
Qed.

(* ************************ *)

Lemma translate_lift_exp :
  forall a m n,
    translate_exp (lift a m n) = lift_texp (translate_exp a) m n.
Proof.
induction a; intros; simpl in *; 
 rewrite_ih; auto.
Qed.
Hint Rewrite translate_lift_exp : step.

Lemma translate_lift_typ :
  forall A th m n,
    translate_typ (lift A m n) th = lift_ttyp (translate_typ A th) m n.
Proof.
induction A; intros; simpl in *;
  autorewrite with step;
  rewrite_ih; auto.
Qed.
Hint Rewrite translate_lift_typ : step.

Lemma translate_open_exp' :
 forall a b i,
   translate_exp ({i ~> lift a i 0}b)
   = {i ~t> lift_texp (translate_exp a) i 0}(translate_exp b).
Proof.
induction b; intros; simpl in *;
  try db_var_case; simpl;
  autorewrite with step;
  rewrite_ih;
  auto.
Qed.
Hint Rewrite translate_open_exp' : step.

Lemma translate_open_exp : 
  forall a b,
    translate_exp ({0 ~> a}b)
   = ({0 ~t> translate_exp a}(translate_exp b)).
Proof.
  intros.
  replace a with (lift a 0 0) by (apply lift_0_is_id).
  rewrite translate_open_exp'.
  rewrite lift_0_is_id.
  rewrite lift_texp_0_is_id.
  reflexivity.
Qed.

Lemma typelike_lift : forall a i j,
  typelike (lift a i j) -> typelike a.
Proof.
induction a; intros; simpl in *; auto.
Qed.

Lemma translate_open_typ' :
  forall a, ~typelike a 
  -> forall B th i,
       translate_typ ({i~>lift a i 0}B) th 
         = {i ~T> lift_texp (translate_exp a) i 0}(translate_typ B th).
Proof.
induction B; intros; simpl in *;
  autorewrite with step;
   rewrite_ih; auto.

  (* TVar case *)
  db_var_case; simpl;
    try rewrite (translate_typ_of_not_typelike); eauto using typelike_lift.
Qed.

Lemma translate_open_typ : forall G th th1 th2 a A B,
  typing G th1 a A 
  -> typing G th2 A exp_star 
  -> translate_typ ({0~>a}B) th 
      = {0 ~T> translate_exp a}(translate_typ B th).
Proof.
  intros.
  assert (translate_typ ({0~>lift a 0 0}B) th 
           = {0 ~T> lift_texp (translate_exp a) 0 0}(translate_typ B th))
    by eauto using translate_open_typ', type_typelike, term_not_typelike.
  autorewrite with step in *.
  auto.
Qed.

Lemma translate_pstep : forall a b,
  pstep a b -> ptstep (translate_exp a) (translate_exp b).
Proof.
  Hint Constructors ptstep.
  induction 1; simpl in *; 
   repeat rewrite translate_open_exp; simpl;
   try rewrite translate_lift_exp;
   auto using translate_val.
Qed.

Lemma translate_mpstep : forall a b,
  mpstep a b -> mptstep (translate_exp a) (translate_exp b).
Proof.
  induction 1; intros;
  eauto using translate_pstep.
Qed.

Lemma translate_better_theta : forall A th1 th2,
     better_theta th1 th2
  -> better_ttyp (translate_typ A th1) (translate_typ A th2).
Proof.
   induction A; simpl; intros; auto.
Qed.

(* Move this to a better place *)
Lemma better_theta_log : forall th,
  better_theta th_log th.
Proof.
  destruct th; auto.
Qed.
Hint Resolve better_theta_log.

(* Move this to a better place *)
Lemma better_theta_prog : forall th,
  better_theta th th_prog.
Proof.
  destruct th; auto.
Qed.
Hint Resolve better_theta_prog.

Lemma translate_better : forall A B,
     better_typ A B 
  -> forall th1 th2,
       better_theta th1 th2
       -> better_ttyp (translate_typ A th1) (translate_typ B th2).
Proof.
  induction 1; 
    inversion 1; subst; simpl; 
    eauto;
    match goal with
     [ H : FO ?A |- better_ttyp (translate_typ ?A ?th1) (translate_typ ?A ?th2) ]
      => rewrite translate_FO with (th:=th1) (th':=th2) by auto; auto
    end.
Qed.

Lemma better_theta_better_typ : forall A th1 th2,
   better_theta th1 th2
   -> better_ttyp (translate_typ A th1) (translate_typ A th2).
Proof.
  induction A; 
    inversion 1; subst; simpl; eauto. 
Qed.

Lemma translate_better_L : forall A B,
     better_typ_L A B 
     -> better_ttyp (translate_typ A th_log) (translate_typ B th_log).
Proof.
  induction 1; simpl;
    eauto using translate_better. 
Qed.

Lemma translate_better_typ : forall A B th,
   better_typ A B
  -> better_ttyp (translate_typ A th) (translate_typ B th).
Proof.
   intros. auto using translate_better.
Qed.


Lemma kinding_at_any_theta' : forall G A,
  tkinding G A
   -> forall C th1 th2,
        A = (translate_typ C th1)
         -> tkinding G (translate_typ C th2).
Proof.
  induction 1;
    induction C; 
       simpl; intros th1 th2 Heq; try congruence;
         try injection Heq; intros; subst;
         eauto.
Qed.

(* I thought this lemma would help with translating 
   termination casts, but it doesn't. *)
Lemma kinding_at_any_theta : forall G A th1 th2,
  tkinding G (translate_typ A th1)
  -> tkinding G (translate_typ A th2).
Proof.
  eauto using kinding_at_any_theta'.
Qed.

(* Combined induction for the typing, kinding, and ctx_wf judgements.
  The "component" theorems are:

  Theorem translate_ctx_wf : forall G,
    ctx_wf G -> tctx_wf (translate_ctx G).

  Theorem translate_kinding : forall G th A A',
      typing G th A A'
   -> A' = exp_star
   -> forall th', tkinding (translate_ctx G) (translate_typ A th').

  Theorem translate_typing : forall G th a A,
      typing G th a A
   -> ttyping (translate_ctx G) th (translate_exp a) (translate_typ A th).
 *)
Theorem translate_combined :
  forall G th a A,
   typing G th a A ->
     ttyping (translate_ctx G) th (translate_exp a) (translate_typ A th)
     /\
     (A = exp_star
       -> forall th', tkinding (translate_ctx G) (translate_typ a th'))
     /\ tctx_wf (translate_ctx G).
Proof.     
  apply (typing_ind_mut
          (fun G th a A _ =>
             ttyping (translate_ctx G) th (translate_exp a) (translate_typ A th)
             /\
             (A = exp_star
               -> forall th', tkinding (translate_ctx G) (translate_typ a th'))
             /\ tctx_wf (translate_ctx G))
          (fun G _ =>  tctx_wf (translate_ctx G)));
  intros;
  repeat match goal with 
   | [ H : _ /\ _ |- _ ] => destruct H
  end;
  auto;
  try (split; [ idtac | split] ;
        [   (* Cases from the typing part: *)
          simpl in *; auto
          (* Cases from the kinding part: *)
        | intros H_is_star th'; try rewrite H_is_star in *;
                        try (elimtype False; now eauto 2 using not_star_star, not_better_star, not_better_L_star);
          auto
        | auto ]).

  (* TVar typing *)
+ rewrite translate_ctx_app in *.
  simpl in *.
  rewrite translate_lift_typ in *.
  apply TTVar.
    auto.
    rewrite <- translate_lift_typ. auto.
    rewrite translate_ctx_length; auto.

  (* TUnboxVal typing *)
+ eauto using value_any_theta, translate_val.

  (* TUnboxL *)
+ destruct t'.
  simpl in *. auto.
  apply subsume_theta with (th:=th_log). auto.

  (* TUnboxP *)
+ eauto using translate_better_theta.
  
  (* TArr kinding *)
+ simpl in *.
  auto.

  (* TApp typing *)
+ go_go_regularity.
  erewrite translate_open_typ by eauto.
  apply TTApp with (th2:=t) (A:=translate_typ A th_log).
    eauto.
    eauto.
    rewrite translate_FO with (th':=t); auto.
    eauto.
    erewrite <- translate_open_typ by eauto; eauto.

  (* TLamP *)
+ go_go_regularity.   
  apply TTLam.
    rewrite translate_FO with (th':=th_prog); auto.
    auto.

  (* TRec typing *)
+ go_go_regularity.
  autorewrite with step in *.
  replace (translate_typ A th_prog) with (translate_typ A th_log)
          in * by (eauto using translate_FO).
   auto.

  (* TNRec typing *)
+ go_go_regularity.
  repeat rewrite translate_lift_typ in *.
  auto.

  (* TSub typing *)
+ eapply TTSubtype.
   eauto using subsume_theta.
   eauto using translate_better_theta.
   eauto.

  (* TSubL typing *)
+ eauto using subsume_theta.

  (* TFOVal typing *)
+ rewrite translate_FO with (th' := th_prog);
  eauto using value_any_theta, translate_val.

  (* TSigma kinding *)
+ simpl in *.
  auto.

  (* TPair typing *)
+ go_go_regularity;
  apply TTPair.
   auto.
   rewrite translate_FO with (th':=t); auto.
   erewrite <- translate_open_typ by eauto; eauto.
   erewrite <- translate_open_typ by eauto; eauto.

  (* TPCase typing *)
+ go_go_regularity.
  apply TTPCase with (A1 := translate_typ A1 t') (A2 := translate_typ A2 t').
   (* scrutinee: *)
   replace (translate_typ A1 t') with (translate_typ A1 th_log)
      by (eauto using translate_FO).
   auto.
   (* body: *)
   rewrite <- translate_lift_typ.
   rewrite <- translate_lift_exp.
   auto.
   (* kinding premise: *)
   auto.

  (* TNCase typing *)
+ apply TTNCase.
    (* scrutinee: *)
    auto.
    (* zero branch *)
    rewrite <- translate_lift_typ.
    auto.
    (* succ branch *)
    rewrite <- translate_lift_typ.
    rewrite <- translate_lift_exp.
    auto.  
    (* kinding premise: *)
    auto.

  (* TEq kinding *)
+ apply TKEq with (A := translate_typ A t1) (t1:=t1) (B := translate_typ B t2) (t2 := t2);
   auto.

  (* TJoin typing *)
+ eapply TTJoin with (c := translate_exp c);
   eauto using translate_mpstep.

  (* TConv typing *)
+ erewrite translate_open_typ in * by eauto.
  apply TTConv with (p := translate_exp p) 
                    (b1 := translate_exp b1)
                    (B1 := translate_typ B1 th_prog)
                    (B2 := translate_typ B2 th_prog);
    auto.
    
    erewrite <- translate_open_typ.
    apply H6.
      auto.    
      eassumption.
      eassumption.

  (* TSubtype typing *)
+ eauto using translate_better_typ.

  (* TSubtype_L typing *)
+   eauto using translate_better_L.

  (* TAt kinding *)
+ simpl. auto.
  
  (* TAtP typing *)
+ eauto using subsume_theta.

  (* TAtL typing  *)
+ eauto using translate_better_theta.

  (* TAtLV  typing *)
+ eauto using value_any_theta, translate_val.

  (* CWFTrm  ctx_wf *)
+ simpl; auto.
Qed.  

Corollary translate_typing : forall G th a A,
      typing G th a A
   -> ttyping (translate_ctx G) th (translate_exp a) (translate_typ A th).
Proof.
  intros T th a A H.
  destruct (translate_combined _ _ _ _ H) as [? [? ?]].
  auto.
Qed.

Print Assumptions translate_typing. 