(* Changes from the "real" lambda-theta definition:

   Changes motivated by step_2_teqt:
    - All rules to do with sum types and recursive types have been omitted.
    - ctx_wf no longer allows type variables in the context (not needed since we don't
      have recursive types).
    - Added a regularity premises to TVar, TSub. Given the above change to ctx_wf this has 
       no effect on what terms are typable, but it gives a better IH.
       (maybe can define a "cleaner" version of typesystem to get rid of it)
    - TConv now has two new premisses forbidding elimination of large equations.
       (In the "cleaner" version these could be moved to the TEq rule, which would
        arguably look prettier).
    - The rules TContra and T*Inv are omitted (they are only interesting if we have
      large equations).

   Changes motivated by teqt_2_step:
    - Added TUnboxP and TUnboxL (used for the typing_unbox lemma which
       is used e.g. in the TTApp case).
    - Added an extra rule TLamP, which let's you use programmatic lambdas like in Teqt.
*)

(* Why do we need those extra regularity premises? E.g. in TSub, we
need to apply subtyping, so we need to known that the desired type in
Teqt is wellformed. By just regularity we'd know that the
corresponding type in lambda-theta is wellformed. But we are proving
translation of wellformedness mutually with translation of typing, so
this is not good enough, we need it as a subderivation in order to get
an IH.  *)

Require Import Arith.
Require Import Bool.
Require Import List.

Require Import shared_defs.

Inductive exp : Set := 
 | exp_var    (n:nat)
 | exp_star
 | exp_arr    (A:exp) (B:exp)
 | exp_fun    (a:exp)
 | exp_lam    (a:exp)
 | exp_nrec   (a:exp)
 | exp_app    (a:exp) (b:exp)
 | exp_at     (A:exp) (t:theta)
 | exp_nat
 | exp_zero   
 | exp_succ   (a:exp)
 | exp_ncase  (a:exp) (b1:exp) (b2:exp)
 | exp_sigma  (A:exp) (B:exp)
 | exp_pair   (a:exp) (b:exp)
 | exp_pcase  (a:exp) (b:exp)
 | exp_eq     (a b:exp)
 | exp_join
(*
 | exp_Terminates (a:exp) (A:exp) 
 | exp_terminates
*)
 .

Definition ctx : Set := list (theta * exp).

(** subrules *)
Inductive val : exp -> Prop :=
  | val_var  : forall n, val (exp_var n)
  | val_star : val exp_star
  | val_arr  : forall A B, val (exp_arr A B)
  | val_fun  : forall a, val (exp_fun a)
  | val_lam  : forall a, val (exp_lam a)
  | val_nrec : forall a, val (exp_nrec a)
  | val_at   : forall A t, val (exp_at A t)
  | val_nat  : val exp_nat
  | val_zero : val exp_zero
  | val_succ : forall v, val v -> val (exp_succ v)
  | val_sigma: forall A B, val (exp_sigma A B)
  | val_pair : forall v u, val v -> val u -> val (exp_pair v u)
  | val_eq   : forall a b, val (exp_eq a b)
  | val_join : val exp_join
(*
  | val_terminates : val exp_terminates
*)
.

(** definitions *)
(* lift e n m  -  adds n to every var which is at least as big as m.

   so if     G1 ++ G3 |- e : A
   and       length G2 = n      length G1 = m
   then 
        (lift_c G1 n m) ++ G2 ++ G3 |- lift e n m : (lift A n m)
 *)
Fixpoint lift (e : exp) (n m : nat) : exp :=
  match e with
  | exp_var n'        => exp_var (match nat_compare n' m with
                                  | Lt => n'
                                  | _  => (n' + n)
                                  end)
  | exp_star          => exp_star
  | exp_arr A1 A2     => exp_arr (lift A1 n m) (lift A2 n (S m))
  | exp_fun  a        => exp_fun  (lift a n (S (S m)))
  | exp_lam  a        => exp_lam  (lift a n (S m))
  | exp_nrec a        => exp_nrec (lift a n (S (S m)))
  | exp_app  a b      => exp_app  (lift a n m) (lift b n m)
  | exp_at A t        => exp_at (lift A n m) t
  | exp_nat           => exp_nat
  | exp_zero          => exp_zero
  | exp_succ a        => exp_succ (lift a n m)
  | exp_ncase a b1 b2 => exp_ncase (lift a n m) (lift b1 n (S m))
                                   (lift b2 n (S (S m)))
  | exp_sigma A B    => exp_sigma (lift A n m) (lift B n (S m))
  | exp_pair a b     => exp_pair  (lift a n m) (lift b n m)
  | exp_pcase a b    => exp_pcase (lift a n m) (lift b n (S (S (S m))))
  | exp_eq a b       => exp_eq (lift a n m) (lift b n m)
  | exp_join         => exp_join
(*
  | exp_Terminates a A => exp_Terminates (lift a n m) (lift A n m)
  | exp_terminates   => exp_terminates
*)
  end.

(* lift_c G1 n m - adds n to the term variables of G1 that refer at least
   m places outside of G1

   so if     |- G1 ++ G3
   and       length G2 = n
   then 
         |- (lift_c G1 n) ++ G2 ++ G3
 *)
Fixpoint lift_c (G : ctx) n m : ctx := 
  match G with
  | nil         => nil
  | (t,A) :: G' => (t,lift A n (m + length G')) :: lift_c G' n m
  end.


Reserved Notation "{ n ~> a }" (at level 60, no associativity).

Fixpoint open (n : nat) (u : exp) (e : exp) {struct e} : exp :=
  match e with
  | exp_var  n'   => 
    match nat_compare n' n with
      | Lt => exp_var n'
      | Eq => u
      | Gt => exp_var (pred n')
    end
  | exp_star      => exp_star
  | exp_arr A B   => exp_arr ({n ~> u}A) ({S n ~> lift u 1 0} B)
  | exp_fun  a    => exp_fun ({S (S n) ~> lift u 2 0} a)
  | exp_lam  a    => exp_lam ({S n ~> lift u 1 0} a)
  | exp_nrec  a   => exp_nrec ({S (S n) ~> lift u 2 0} a)
  | exp_app  a b  => exp_app ({n ~> u}a) ({n ~> u}b)
  | exp_at A t    => exp_at ({n ~> u} A) t
  | exp_nat       => exp_nat
  | exp_zero      => exp_zero
  | exp_succ a    => exp_succ ({n ~> u} a)
  | exp_ncase a b1 b2 => exp_ncase ({n ~> u} a) ({S n ~> lift u 1 0}b1) 
                                   ({S (S n) ~> lift u 2 0} b2)
  | exp_sigma A B => exp_sigma ({n ~> u} A) ({S n ~> lift u 1 0}B)
  | exp_pair a b  => exp_pair ({n ~> u} a) ({ n ~> u}b)
  | exp_pcase a b => exp_pcase ({n ~> u} a) ({(S (S (S n))) ~> lift u 3 0}b)
  | exp_eq a b    => exp_eq ({n ~> u}a) ({n ~> u}b)
  | exp_join      => exp_join
(*
  | exp_Terminates a A  => exp_Terminates ({n ~> u} a) ({n ~> u} A)
  | exp_terminates => exp_terminates
*)
  end
  where "{ n ~> a }" := (open n a).

Fixpoint open_c (n : nat) (u : exp) (G : ctx) : ctx := 
  match G with
  | nil     => nil
  | (t,A) :: G' =>
       (t,{n + length G' ~> lift u (n + length G') 0} A) :: (open_c n u G')
  end.

(* count the number of free variables *)
Fixpoint fvs (a : exp) : nat :=
  match a with
  | exp_var n   => S n
  | exp_star     => 0
  | exp_arr A B => max (fvs A) (pred (fvs B))
  | exp_fun a   => pred (pred (fvs a))
  | exp_lam a   => pred (fvs a)
  | exp_nrec a  => pred (pred (fvs a))
  | exp_app a b => max (fvs a) (fvs b)
  | exp_at A _  => fvs A
  | exp_nat     => 0
  | exp_zero    => 0
  | exp_succ a  => fvs a
  | exp_ncase a b1 b2 => max (fvs a) (max (pred (fvs b1))
                                          (pred (pred (fvs b2))))
  | exp_sigma A B => max (fvs A) (pred (fvs B))
  | exp_pair a b  => max (fvs a) (fvs b)
  | exp_pcase a b => max (fvs a) (pred (pred (pred (fvs b))))
  | exp_eq a b    => max (fvs a) (fvs b)
  | exp_join      => 0
(*
  | exp_Terminates a A => max (fvs a) (fvs A)
  | exp_terminates => 0
*)
  end.
 
(* defns JStepCBV *)
Inductive step : exp -> exp -> Prop :=    (* defn step *)
 | SCaseZ : forall (a1 a2:exp),
     step (exp_ncase exp_zero a1 a2) ({0 ~> exp_join} a1)
 | SCaseS : forall (v a1 a2:exp),
     val v ->
     step (exp_ncase (exp_succ v) a1 a2)  
          ({0 ~> v} ({0 ~> exp_join} a2))
 | SCasePair : forall v1 v2 b,
     val v1 ->
     val v2 ->
     step (exp_pcase (exp_pair v1 v2) b)
          ({0 ~> v1} ({0 ~> lift v2 1 0} ({0~> exp_join}b)))
 | SBetaFun : forall (a v:exp),
     val v ->
     step (exp_app (exp_fun a) v)
          ({0 ~> v} ({0 ~> exp_fun (lift a 1 2)} a))
 | SBetaLam : forall (a v:exp),
     val v ->
     step (exp_app (exp_lam a) v) ({0 ~> v} a) 
 | SBetaNRec : forall (a v:exp),
     val v ->
     step (exp_app (exp_nrec a) v)
          ({0 ~> v} 
           ({0 ~> exp_lam (exp_lam (exp_app 
                     (exp_nrec (lift a 3 2)) (exp_var 1)))} a))
 | SSucc1 : forall a a',
     step a a' ->
     step (exp_succ a) (exp_succ a')
 | SApp1 : forall (a a' b : exp),
     step a a' -> 
     step (exp_app a b) (exp_app a' b)
 | SApp2 : forall (v b b' : exp),
     val v -> 
     step b b' -> 
     step (exp_app v b) (exp_app v b')
 | SNCase1 : forall (a a' b1 b2 : exp),
     step a a' -> 
     step (exp_ncase a b1 b2) (exp_ncase a' b1 b2)
 | SPair1 : forall a a' b,
     step a a' ->
     step (exp_pair a b) (exp_pair a' b)
 | SPair2 : forall v b b',
     val v ->
     step b b' ->
     step (exp_pair v b) (exp_pair v b')
 | SPCase1 : forall a a' b,
     step a a' ->
     step (exp_pcase a b) (exp_pcase a' b).


(** definitions *)
(* defns JMultiStep *)
Inductive mstep : nat -> exp -> exp -> Prop :=    (* defn mstep *)
 | MSRefl : forall (a:exp),
     mstep 0 a a
 | MSStep : forall (n:nat) (a:exp) (b' b:exp),
     step a b -> 
     mstep n b b' -> 
     mstep (S n) a b'.
(** definitions *)

(*******************************************************************)

Inductive pstep : exp -> exp -> Prop :=
 | PRefl : forall a, pstep a a
 | PCaseZ : forall (a1 a1' a2:exp),
      pstep a1 a1'
   -> pstep (exp_ncase exp_zero a1 a2) ({0 ~> exp_join} a1')
 | PCaseS : forall (v v' a1 a2 a2':exp),
      val v
   -> pstep v v'
   -> pstep a2 a2'
   -> pstep (exp_ncase (exp_succ v) a1 a2)
            ({0 ~> v'} ({0 ~> exp_join} a2'))
 | PCasePair : forall  v1 v1' v2 v2' b b',
      val v1 
   -> pstep v1 v1'
   -> val v2
   -> pstep v2 v2'
   -> pstep b b'
   -> pstep (exp_pcase (exp_pair v1 v2) b)
            ({0 ~> v1'} ({0 ~> lift v2' 1 0} ({0 ~> exp_join} b')))
 | PBetaFun : forall (a a' v v':exp),
      val v 
   -> pstep v v'
   -> pstep a a'
   -> pstep (exp_app (exp_fun a) v)
            ({0 ~> v'} ({0 ~> exp_fun (lift a' 1 2)} a'))
 | PBetaLam : forall (a a' v v':exp),
      val v 
   -> pstep v v'
   -> pstep a a'
   -> pstep (exp_app (exp_lam a) v) ({0 ~> v'} a')
 | PBetaNRec : forall (a a' v v':exp),
      val v 
   -> pstep v v'
   -> pstep a a'
   -> pstep (exp_app (exp_nrec a) v)
            ({0 ~> v'}
             ({0 ~> exp_lam (exp_lam (exp_app
                      (exp_nrec (lift a' 3 2)) (exp_var 1)))} a'))
 | PFun1 : forall (a a':exp),
      pstep a a'
   -> pstep (exp_fun a) (exp_fun a')
 | PLam1 : forall (a a':exp),
      pstep a a'
   -> pstep (exp_lam a) (exp_lam a')
 | PNRec1 : forall (a a':exp),
      pstep a a'
   -> pstep (exp_nrec a) (exp_nrec a')
 | PSucc1 : forall a a',
      pstep a a'
   -> pstep (exp_succ a) (exp_succ a')
 | PApp1: forall (a a' b b' : exp),
      pstep a a'
   -> pstep b b'
   -> pstep (exp_app a b) (exp_app a' b')
 | PNCase1 : forall (a a' b1 b1' b2 b2' : exp),
      pstep a a'
   -> pstep b1 b1'
   -> pstep b2 b2'
   -> pstep (exp_ncase a b1 b2) (exp_ncase a' b1' b2')
 | PPCase1 : forall a a' b b',
      pstep a a'
   -> pstep b b'
   -> pstep (exp_pcase a b) (exp_pcase a' b')
 | PPair1 : forall (a a' b b' : exp),
      pstep a a'
   -> pstep b b'
   -> pstep (exp_pair a b) (exp_pair a' b')
 | PArr1 : forall A A' B B',
      pstep A A'
   -> pstep B B'
   -> pstep (exp_arr A B) (exp_arr A' B')
 | PAt1 : forall A A' t,
      pstep A A'
   -> pstep (exp_at A t) (exp_at A' t)
 | PSigma1 : forall A A' B B',
      pstep A A'
   -> pstep B B'
   -> pstep (exp_sigma A B) (exp_sigma A' B')
 | PEq1 : forall a a' b b',
      pstep a a'
   -> pstep b b'
   -> pstep (exp_eq a b) (exp_eq a' b')
(*
 | PTerminates1 : forall a a' A A',
      pstep a a'
   -> pstep A A'
   -> pstep (exp_Terminates a A) (exp_Terminates a' A')
*)
.

Inductive mpstep : exp -> exp -> Prop :=
  | MPRefl : forall a, mpstep a a
  | MPStep : forall a b c, pstep a b -> mpstep b c -> mpstep a c.

Inductive FO : exp -> Prop :=
 | FONat : 
     FO exp_nat
 | FOEq : forall a b,
     FO (exp_eq a b)
 | FOAt : forall (A:exp) (t:theta),
     FO (exp_at A t)
 | FOSigma : forall A B,
     FO A -> FO B -> FO (exp_sigma A B)
(*
 | FOTerminates : forall a A,
     FO (exp_Terminates a A)
*)
.

Inductive head : Set :=
 | head_star
 | head_arr | head_at (t : theta) | head_nat | head_sum | head_sigma | head_mu | head_eq (* | head_Terminates *) .

Inductive head_form : exp -> head -> Prop :=
  | HStar : head_form exp_star head_star
  | HArr : forall A B, head_form (exp_arr A B) head_arr
  | HAt : forall A t, head_form (exp_at A t) (head_at t)
  | HNat : head_form exp_nat head_nat
  | HSigma : forall A B, head_form (exp_sigma A B) head_sigma
  | HEq : forall a b, head_form (exp_eq a b) head_eq
(*
  | HTerminates : forall a A, head_form (exp_Terminates a A) head_Terminates
*)
  .

(* Subtyping *)
Inductive better_typ : exp -> exp -> Prop :=
  | BeTAt : forall A A' th th',
       better_theta th th'
    -> better_typ A A'
    -> better_typ (exp_at A th) (exp_at A' th')
  | BeTAtP : forall A A',
       better_typ A A'
    -> better_typ A (exp_at A' th_prog)
  | BeTAtL : forall A A',
       better_typ A A'
    -> better_typ (exp_at A th_log) A'
  | BeTAtFO1 : forall A th,
       FO A
    -> better_typ A (exp_at A th)
  | BeTAtFO2 : forall A th,
       FO A
    -> better_typ (exp_at A th) A
  | BeTArr : forall A A' B B',
       better_typ A' A
    -> better_typ B B'
    -> better_typ (exp_arr A B) (exp_arr  A' B')
  | BeTSigma : forall A B B',
    better_typ B B'
    -> better_typ (exp_sigma A B) (exp_sigma A B')
  | BeTNat : better_typ exp_nat exp_nat
  | BeTEq : forall a b, better_typ (exp_eq a b) (exp_eq a b)
(*
  | BeTTerminates : forall a A, better_typ (exp_Terminates a A) (exp_Terminates a A)
*)
.

(* This is pretty hacky... *)
Inductive better_typ_L : exp -> exp -> Prop :=
  | BeTLBetter : forall A A',
       better_typ A A'
    -> better_typ_L A A'
  | BeTLAtL : forall A A',
       better_typ_L A A'
    -> better_typ_L A (exp_at A' th_log)
  | BeTLArr : forall A A' B B',
       better_typ A' A
    -> better_typ_L B B'
    -> better_typ_L (exp_arr A B) (exp_arr  A' B').


Inductive ctx_wf : ctx -> Prop :=
 | CWFNil  : ctx_wf nil
(*  | CWFTyp : forall t G, ctx_wf G -> ctx_wf ((t,exp_star)::G) *)
 | CWFTrm : forall t A G,
       typing G th_log A exp_star
    -> ctx_wf G
    -> ctx_wf ((t,A) :: G)

with typing : ctx -> theta -> exp -> exp -> Prop :=
(* | TContra : forall G t t' p B1 B2 h1 h2 a A B,
        typing G th_log p (exp_eq B1 B2)
     -> head_form B1 h1 -> head_form B2 h2
     -> h1 <> h2
     -> typing G t a A
     -> typing G t A exp_star
     -> typing G t' B exp_star
     -> typing G t' a B
 | TEqAtInv : forall G t p A B t',
        typing G t p (exp_eq (exp_at A t') (exp_at B t'))
     -> typing G t (exp_eq A B) exp_star
     -> typing G t p (exp_eq A B)
 | TEqArrInv1 : forall G t p A1 B1 A2 B2,
        typing G t p (exp_eq (exp_arr A1 B1) (exp_arr A2 B2))
     -> typing G t (exp_eq A1 A2) exp_star
     -> typing G t p (exp_eq A1 A2)
 | TEqArrInv2 : forall G t t' v p A1 B1 A2 B2,
        typing G t p (exp_eq (exp_arr A1 B1) (exp_arr A2 B2))
     -> typing G t' v A1
     -> val v
     -> typing G t (exp_eq ({0 ~> v}B1) ({0 ~> v}B2)) exp_star
     -> typing G t p (exp_eq ({0 ~> v}B1) ({0 ~> v}B2))
 | TEqSigmaInv1 : forall G t p A1 B1 A2 B2,
        typing G t p (exp_eq (exp_sigma A1 B1) (exp_sigma A2 B2))
     -> typing G t (exp_eq A1 A2) exp_star
     -> typing G t p (exp_eq A1 A2)
 | TEqSigmaInv2 : forall G t t' v p A1 B1 A2 B2,
        typing G t p (exp_eq (exp_sigma A1 B1) (exp_sigma A2 B2))
     -> typing G t' v A1
     -> val v
     -> typing G t (exp_eq ({0 ~> v}B1) ({0 ~> v}B2)) exp_star
     -> typing G t p (exp_eq ({0 ~> v}B1) ({0 ~> v}B2))            *)
 | TVar : forall (G1 G3:ctx) (t:theta) (n:nat) (A:exp),
     ctx_wf (G1 ++ (t,A) :: G3) ->
     n = length G1 ->
     typing (G1 ++ (t,A) :: G3) th_log (lift A (S n) 0) exp_star ->
     typing (G1 ++ (t,A) :: G3) t (exp_var n) (lift A (S n) 0)
 | TUnboxVal : forall G t v A t',
       typing G t  v (exp_at A t')
    -> val v
    -> typing G th_log A exp_star
    -> typing G t' v A
 | TUnboxL : forall G a A t',
       typing G th_log a (exp_at A t')
    -> typing G th_log A exp_star
    -> typing G t' a A
 | TUnboxP : forall G a A t',
       typing G th_prog a (exp_at A t')
    -> typing G th_log A exp_star
    -> typing G th_prog a A
 | TArr : forall G A B,
       typing G th_log A exp_star
    -> FO A
       (* It may look strange that B can assume x to be logical,
          but since A is FO it doesn't matter. *)
    -> typing ((th_log,A) :: G) th_log B exp_star
    -> typing G th_log (exp_arr A B) exp_star
 | TApp : forall (G:ctx) (t:theta) (a b A B:exp),
       typing G t a (exp_arr A B)
    -> typing G t b A
    -> typing G th_log ({0 ~> b} B) exp_star
    -> typing G t (exp_app a b) ({0 ~> b} B)
 | TLam : forall (G:ctx) (b A B:exp),
       typing ((th_log,A) :: G) th_log b B
    -> typing G th_log (exp_arr A B) exp_star
    -> typing G th_log (exp_lam b) (exp_arr A B)
 | TLamP : forall (G:ctx) (b A B:exp),
       typing ((th_prog,A) :: G) th_prog b B
    -> typing G th_log (exp_arr A B) exp_star
    -> typing G th_prog (exp_lam b) (exp_arr A B)
 | TRec : forall (G:ctx) (a A B:exp),
     typing ((th_prog,lift (exp_arr A B) 1 0) :: (th_prog,A) :: G)
            th_prog a (lift B 1 0) ->
     typing G th_log (exp_arr A B) exp_star ->
     typing G th_prog (exp_fun a) (exp_arr A B)
 | TNRec : forall (G:ctx) (a B:exp),
     typing (   (th_log,exp_arr exp_nat 
                     (exp_arr (exp_eq (exp_succ (exp_var 0))
                                     (exp_var 1))
                              (lift (lift B 1 1) 1 0)))
             :: (th_log,exp_nat) :: G)
            th_log a (lift B 1 0) ->
     typing G th_log (exp_arr exp_nat B) exp_star ->
     typing G th_log (exp_nrec a) (exp_arr exp_nat B)
(* By regularity, both TSub and TSubType could be subsumed by a single rule which omits the
   wellformedness premise. *)
 | TSub : forall (G:ctx) (a A:exp),
     typing G th_log a A ->
     typing G th_log A exp_star ->
     typing G th_prog a A
 | TSubType : forall (G:ctx) (A:exp),
     typing G th_log A exp_star ->
     typing G th_prog A exp_star
 | TFOVal : forall (G:ctx) (v A:exp),
       val v
    -> FO A
    -> typing G th_prog v A
    -> typing G th_log A exp_star
    -> typing G th_log v A
 | TNat : forall G,
     ctx_wf G -> typing G th_log exp_nat exp_star
 | TZero : forall (G:ctx) (t:theta),
     ctx_wf G ->
     typing G t exp_zero exp_nat
 | TSuc : forall (G:ctx) (t:theta) (b:exp),
     typing G t b exp_nat ->
     typing G t (exp_succ b) exp_nat

 | TSigma : forall G A B,
      typing G th_log A exp_star
   -> FO A
   -> typing ((th_log,A) :: G) th_log B exp_star
   -> typing G th_log (exp_sigma A B) exp_star

 | TPair : forall G t A B a b,
      typing G th_log (exp_sigma A B) exp_star
   -> typing G t a A
   -> typing G th_log ({0 ~> a}B) exp_star
   -> typing G t b ({0~>a}B)
   -> typing G t (exp_pair a b) (exp_sigma A B)

 | TPCase : forall G t t' A1 A2 B a b,
      typing G t a (exp_at (exp_sigma A1 A2) t') 
   -> typing (   (th_log, exp_eq (exp_pair (exp_var 1) (exp_var 0)) (lift a 2 0))
              :: (t', A2) 
              :: (t', A1)
              :: G)
             t b (lift B 3 0)
   -> typing G th_log B exp_star
   -> typing G t (exp_pcase a b) B

 | TNCase : forall (G:ctx) (t:theta) (a b1 b2 A:exp),
     typing G t a exp_nat ->
     typing ((th_log,exp_eq exp_zero a) :: G) t b1 (lift A 1 0) ->
     typing (   (th_log,exp_eq (exp_succ (exp_var 0)) (lift a 1 0)) 
             :: (t,exp_nat) :: G)
            t b2 (lift A 2 0) ->
     typing G th_log A exp_star ->
     typing G t (exp_ncase a b1 b2) A
(* For these two rules, we could pick either t1/t2 or th_prog, but the 
   choice should be consistent. Otherwise we will have trouble proving
   regularity--because TSub now has a regularity assumption. *) 
 | TEq  : forall G a b t1 t2 A B,
       typing G t1 a A -> typing G t2 b B
    -> typing G th_log A exp_star -> typing G th_log B exp_star
    -> typing G th_log (exp_eq a b) exp_star
 | TJoin : forall G a b c t1 t2 A B,
       mpstep a c -> mpstep b c
    -> typing G t1 a A -> typing G t2 b B
    -> typing G th_log A exp_star -> typing G th_log B exp_star
    -> typing G th_log exp_join (exp_eq a b)
 | TConv : forall G p a b1 b2 B1 B2 t A,
       typing G th_log p (exp_eq b1 b2)
    -> typing G th_prog b1 B1 -> typing G th_log B1 exp_star
    -> typing G th_prog b2 B2 -> typing G th_log B2 exp_star
    -> typing G t a ({0 ~> b1}A)
    -> typing G th_log ({0 ~> b2}A) exp_star
    -> typing G t a ({0 ~> b2}A)
(*
 | TTerminates : forall G a A t,
       typing G th_prog a A
    -> typing G th_log A exp_star
    -> typing G th_log (exp_Terminates a A) exp_star
 | TReify : forall G a A th,
       typing G th_log a A
    -> typing G th exp_terminates (exp_Terminates a A)
 | TReflect : forall G a A th p,
       typing G th_prog a A
    -> typing G th_log p (exp_Terminates a A)
    -> typing G th a A
*)

 | TSubtype : forall G th a A A',
       typing G th a A
    -> better_typ A A'
    -> typing G th_log A' exp_star
    -> typing G th a A'
 | TSubtype_L : forall G a A A',
       typing G th_log a A
    -> better_typ_L A A'
    -> typing G th_log A' exp_star
    -> typing G th_log a A'

 | TAt  : forall G A t t',
       typing G th_log A exp_star
    -> typing G t (exp_at A t') exp_star
 | TAtP : forall (G:ctx) (t:theta) (a A:exp),
       typing G t a A
    -> typing G th_log A exp_star
    -> typing G th_prog a (exp_at A t)
 | TAtL : forall (G:ctx) (t:theta) (a A:exp),
       typing G th_log a A
    -> typing G th_log A exp_star
    -> typing G th_log a (exp_at A t)
 | TAtLV : forall (G:ctx) (v A:exp),
       val v
    -> typing G th_prog v A
    -> typing G th_log A exp_star
    -> typing G th_log v (exp_at A th_prog).

Scheme typing_ind_mut := Induction for typing Sort Prop
  with ctx_wf_ind_mut := Induction for ctx_wf Sort Prop.

Hint Constructors head_form.
Hint Constructors val step mstep FO ctx_wf.
Hint Constructors pstep mpstep.
Hint Resolve TVar TApp TLam TRec TZero TSuc TAtL
             TAtLV TAtP TSub TSubType
             TArr TEq TNat TPair TSigma.
Hint Extern 2 =>
  match goal with
  | [ |- typing _ _ (exp_ncase _ _ _) _ ] => eapply TNCase
  | [ H : typing _ _ ?v (exp_at ?A _)
     |- typing _ _ ?v ?A ] => eapply TUnboxVal
(*
  | [ H : typing _ _ ?v (exp_eq (exp_at _ ?t) (exp_at _ ?t))
     |- typing _ _ ?v (exp_eq _ _) ] => eapply TEqAtInv

  | [ H : typing _ _ ?v (exp_eq (exp_arr ?A1 _) (exp_arr ?A2 _))
     |- typing _ _ ?v (exp_eq ?A1 ?A2) ] => eapply TEqArrInv1

  | [ H : typing _ _ ?v (exp_eq (exp_arr ?A1 _) (exp_arr ?A2 _))
     |- typing _ _ ?v (exp_eq ({_ ~> _}?A1) _) ] => eapply TEqArrInv1

  | [ H : typing _ _ ?v (exp_eq (exp_arr ?A1 _) (exp_arr ?A2 _))
     |- typing _ _ ?v (exp_eq (lift ?A1 _ _) ?A2) ] => eapply TEqArrInv1

  | [ H : typing _ _ ?v (exp_eq (exp_arr _ ?A1) (exp_arr _ _))
     |- typing _ _ ?v (exp_eq ?A1 _) ] => eapply TEqArrInv2

  | [ H : typing _ _ ?v (exp_eq (exp_arr _ ?A1) (exp_arr _ _))
     |- typing _ _ ?v (exp_eq ({_ ~> _}?A1) _) ] => eapply TEqArrInv2

  | [ H : typing _ _ ?v (exp_eq (exp_arr _ ?A1) (exp_arr _ _))
     |- typing _ _ ?v (exp_eq (lift ?A1 _ _) _) ] => eapply TEqArrInv2 *)

  | [ |- typing _ _ (exp_pcase _ _) _ ] => eapply TPCase

(*
  | [ H : typing _ _ ?v (exp_eq (exp_sigma ?A1 _) (exp_sigma ?A2 _))
     |- typing _ _ ?v (exp_eq ?A1 ?A2) ] => eapply TEqSigmaInv1

  | [ H : typing _ _ ?v (exp_eq (exp_sigma _ ?A1) (exp_sigma _ _))
     |- typing _ _ ?v (exp_eq ?A1 _) ] => eapply TEqSigmaInv2

  | [ H : typing _ _ ?v (exp_eq (exp_sigma _ ?A1) (exp_sigma _ _))
     |- typing _ _ ?v (exp_eq ({_ ~> _}?A1) _) ] => eapply TEqSigmaInv2

  | [ H : typing _ _ ?v (exp_eq (exp_sigma _ ?A1) (exp_sigma _ _))
     |- typing _ _ ?v (exp_eq (lift ?A1 _ _) _) ] => eapply TEqSigmaInv2 *)

 | [ |- typing ?G _ (exp_at ?A _) _ ] =>
    first [ apply (TAt G A th_log) | apply TAt ]
  end.

Hint Immediate TSub.


Hint Extern 1 =>
  match goal with
  | [ HS1 : mpstep ?a ?c', HS2 : mpstep ?b ?c'
     |- typing _ _ exp_join (exp_eq (?P ?a) (?P ?b)) ] =>
    eapply TJoin with (c := P c')
  | [ HS1 : mpstep ?a ?c', HS2 : mpstep ?b ?c'
     |- typing _ _ exp_join (exp_eq (?P ?a ?x ?y) (?P ?b ?x ?y)) ] =>
    eapply TJoin with (c := P c' x y)
  | [ HS1 : mpstep ?a ?c', HS2 : mpstep ?b ?c'
     |- typing _ _ exp_join (exp_eq ?a ?b) ] =>
    eapply TJoin with (c := c')
  end.
    
Hint Extern 2 =>
  match goal with
  | [ HFO : FO ?A |- typing _ th_log _ ?A ] =>
    apply TFOVal
  | [ HFO : FO ?A |- typing _ th_log _ (lift ?A _ _) ] =>
    apply TFOVal
  | [ HFO : FO ?A |- typing _ th_log _ ({_ ~> _}?A) ] =>
    apply TFOVal
  | [ HFO : FO ?A |- typing _ th_log _ ({_ ~> _}?A) ] =>
    apply TFOVal
  | [ |- typing _ th_log _ exp_nat ] =>
    apply TFOVal
  | [ |- typing _ th_log _ (exp_at _ _)] =>
    apply TFOVal
  end.

(* Aaron's magic if/then/else *)
Tactic Notation
  "if" tactic(t)
  "then" tactic(t1)
  "else" tactic(t2) :=
  first [ t; first [ t1 | fail 2 ] | t2 ].

(* This builds on if/then/else.  You can use it to demand that two
   metavariables in an Ltac match don't get bound to the same
   thing. *)
Tactic Notation "require_neq" constr(a) constr(b) :=
  if constr_eq a b then fail 2 else idtac.
