Require Import List.

Require Import shared_tactics.
Require Import shared_defs.
Require Import step_defs.
Require Import step_theory.
Require Import teqt_defs.
Require Import teqt_db_basics.

Fixpoint translate_texp (t : texp) : exp :=
  match t with
    | texp_var n => exp_var n
    | texp_fun a => exp_fun (translate_texp a)
    | texp_lam a => exp_lam (translate_texp a)
    | texp_nrec a => exp_nrec (translate_texp a)
    | texp_app a b => exp_app (translate_texp a) (translate_texp b)
    | texp_zero => exp_zero
    | texp_succ a => exp_succ (translate_texp a)
    | texp_ncase a b1 b2 => exp_ncase (translate_texp a) (translate_texp b1) (translate_texp b2)
    | texp_pair a b => exp_pair (translate_texp a) (translate_texp b)
    | texp_pcase a b => exp_pcase (translate_texp a) (translate_texp b)
    | texp_join => exp_join
(*
    | texp_terminates a => exp_let (translate_texp a) (exp_pair (exp_var 0) (exp_var 1))
*)
  end.

Fixpoint translate_ttyp (T : ttyp) : exp :=
  match T with
    | ttyp_nat => exp_nat
    | ttyp_arr th A B => exp_at (exp_arr (translate_ttyp A) (translate_ttyp B)) th
    | ttyp_eq a b => exp_eq (translate_texp a) (translate_texp b)
    | ttyp_sigma A B => exp_sigma (translate_ttyp A) (translate_ttyp B)
(*
    | ttyp_terminates a A => exp_sigma (translate_ttyp A)
                                       (exp_eq (lift (translate_texp a) 1 0)
                                               (exp_var 0))
*)
  end.

Lemma ttyp_FO : forall T, 
  FO (translate_ttyp T).
Proof.
  induction T; simpl; auto.
Qed.

Definition translate_tctx : tctx -> ctx :=
  map (fun A => (th_log, translate_ttyp A)).

Lemma translate_tctx_app : forall G G',
  translate_tctx (G++G') = translate_tctx G ++ translate_tctx G'.
Proof.
  induction G;
     simpl; intros; rewrite_ih; auto.
Qed.

Lemma translate_tctx_length : forall G,
  length (translate_tctx G) = length G.
Proof.
  induction G;
    simpl; auto with arith.
Qed.

Lemma translate_lift_texp : forall a n m,
 translate_texp (lift_texp a n m) = lift (translate_texp a) n m.
Proof.
  induction a; simpl; intros;
    db_var_case; rewrite_ih; auto.
Qed.
Hint Rewrite translate_lift_texp : step.

Lemma translate_lift_ttyp : forall A n m,
   translate_ttyp (lift_ttyp A n m) = lift (translate_ttyp A) n m.
Proof.
  induction A; simpl; intros; rewrite_ih; 
  repeat rewrite translate_lift_texp; auto.
Qed.

Lemma translate_open_texp : forall a b i,
  translate_texp ({i ~t> b}a) = {i ~> translate_texp b}(translate_texp a).
Proof.
  induction a; simpl; intros; rewrite_ih; autorewrite with step; try db_var_case; auto.
Qed.
Hint Rewrite translate_open_texp : step.

(* This lemma is adapted from Chris's step_db_basics.v. *)
Theorem open_lift_commute : forall b a n1 n2 i m,
     m <= n2
  -> m <= i 
  -> {n1 + i ~> (lift a (n1 + n2) 0)} (lift b n1 m) 
   = lift ({i ~> lift a n2 0} b) n1 m.
Proof.
  induction b; intros; simpl in *; eauto;
  autorewrite with step; simpl in *;
  repeat (match goal with
  | [ |- context [ S (_ + _) ] ] =>
    repeat (rewrite plus_n_Sm)
  | [ |- context [ (_ + _ + _) ] ] =>
    repeat (rewrite <- plus_assoc)
  end); rewrite_ih; auto; try db_var_case.
Qed.

Lemma translate_open_ttyp : forall A a i,
  translate_ttyp ({i ~T> a}A) = {i ~> translate_texp a}(translate_ttyp A).
Proof.
  induction A; simpl; intros; auto;
    rewrite_ih; autorewrite with step; auto.
Qed.

Lemma typ_at_inversion1 : forall G t A B,
     typing G t A B
  ->  forall A1 t',
      A = (exp_at A1 t') 
      -> B = exp_star.
Proof.
  induction 1; intros ? ? Heq; try congruence;
    try assert (IH := IHtyping1 _ _ Heq);
    try assert (IH := IHtyping _ _ Heq);
    try assert (IH := IHtyping6 _ _ Heq);
    try congruence;
    subst;
    try match goal with 
      | [ H : better_typ exp_star _ |- _ ] => 
        elimtype False; now eauto using not_better_star2
      | [ H : better_typ_L exp_star _ |- _ ] => 
        elimtype False; now eauto using not_better_L_star2
      | [ H : typing _ _ exp_star _ |- _ ] =>
         elimtype False; now eauto using star_not_typable
    end;
    eauto.

(* The TConv case *)
    destruct A; try destruct n; simpl in *; try congruence.
      subst; elimtype False. 
      eauto using star_not_typable.
Qed.

Lemma typ_at_inversion2 : forall G t A B,
     typing G t A B
  ->  forall A1 t',
      A = (exp_at A1 t') 
      -> typing  G th_log A1 exp_star.
Proof.
  induction 1; intros; try congruence;  eauto.
Qed.

Theorem typ_at_inversion : forall G t A1 t' B,
     typing G t (exp_at A1 t') B
  ->  B = exp_star
      /\ typing G th_log A1 exp_star.
Proof.
  eauto using typ_at_inversion1, typ_at_inversion2.
Qed.

Lemma typing_unbox : forall G th th' a A,
    typing G th a (exp_at A th') 
 -> better_theta th' th
 -> typing G th a A.
Proof.
  intros G th th' a A Htyping Hbetter.
  go_go_regularity;
  apply typ_at_inversion in HReg; destruct HReg. 
  
  inversion Hbetter; subst; [idtac | destruct th];
  eauto using TUnboxP, TUnboxL.
Qed.

Hint Constructors typing.

Lemma typing_boxV : forall G th1 th2 v A,
    typing G th1 v A
 -> val v
 -> typing G th_log A exp_star
 -> typing G th2 v (exp_at A th1).
Proof.
  intros; destruct th1; destruct th2; eauto.
Qed.

Lemma TLamGen : forall (G:ctx) th (b A B:exp),
                  typing ((th,A) :: G) th b B
    -> typing G th_log (exp_arr A B) exp_star
    -> typing G th (exp_lam b) (exp_arr A B).
Proof.
  intros. destruct th; [apply TLam | apply TLamP]; auto.
Qed.

(* Extend a relation pointwise, plus reflexivity at points *)
Inductive onList {A : Type} (P : A -> A -> Prop) : list A -> list A -> Prop :=
  | OLnil   : onList P nil nil
  | OLcons  : forall x xs ys, onList P xs ys -> onList P (x::xs) (x::ys)
  | OLconsP : forall x y xs ys, 
                P x y -> onList P xs ys -> onList P (x::xs) (y::ys).
Hint Constructors onList.

Lemma onList_refl : forall A P (xs : list A),
  onList P xs xs.
Proof.
  induction xs; auto.
Qed.
Hint Resolve onList_refl.

Lemma onList_elt : forall {A} (P : A->A->Prop) G1 x G2 G',
  onList P (G1 ++ x :: G2) G' ->
   exists (G1' : list A) (x':A) (G2':list A),
     G' = (G1' ++ x' :: G2')
     /\length G1 = length G1'
     /\ (x=x' \/ P x x').
Proof.
  induction G1; simpl; intros. 
    + inversion H as [|? ? G2'| ? x' ? G2'].
        - exists nil. exists x.  exists G2'. auto.
        - exists nil. exists x'. exists G2'. auto.
    + inversion H as [|? ? G''| ? a' ? G''].
        - destruct (IHG1 x G2 G'' H3) as [G1' [x' [G2' [? [? [?|?]]]]]].
          subst.
            * exists (a::G1'). exists x'. exists G2'. simpl. auto.
            * exists (a::G1'). exists x'. exists G2'. simpl. 
                split.
                  congruence.
                  split. congruence. auto.
        - destruct (IHG1 x G2 G'' H4) as [G1' [x' [G2' [? [? [?|?]]]]]].
          subst.
            * exists (a'::G1'). exists x'. exists G2'. simpl. auto.
            * exists (a'::G1'). exists x'. exists G2'. simpl. 
                split.
                  congruence.
                  split. congruence. auto.
Qed.              

Lemma strengthening_elementwise : 
  forall (P : (theta * exp) -> (theta * exp) -> Prop)
         (Hstronger : forall th A th' A',
                        P (th,A) (th',A') ->
                        (forall G v n m, val v ->
                          typing G th_log (lift A n m) exp_star
                          -> typing G th' v (lift A' n m) 
                          -> typing G th v (lift A n m))
                        /\ 
                        (forall G n m, 
                              typing G th_log (lift A n m)  exp_star 
                          <-> typing G th_log (lift A' n m) exp_star))
         G th a A
         (Htyping: typing G th a A),
  forall G',  onList P G G' -> typing G' th a A.
Proof.
intros P Hstronger.
  apply (typing_ind_mut
          (fun G th a A _ =>
                     forall G', onList P G G' -> typing G' th a A)
          (fun G _ =>
                     forall G', onList P G G' -> ctx_wf G'));
intros; auto.


(* TVar. This is the only interesting case! *)
+ destruct (onList_elt P _ _ _ _ H1) as [G1' [x' [G2' [? [? Hrelated]]]]].
  destruct Hrelated.
   - subst.
     auto.
   - destruct x' as [t' A'].
     subst.
     destruct (Hstronger t A t' A' H4) as [Hstronger_term Hstronger_type].
     apply Hstronger_term; auto.
       apply TVar.
         auto.
         auto.
         rewrite  <- Hstronger_type; auto.
            
(* TUnboxP *)
+ apply TUnboxP with (t' := t'); auto.

(* TApp *)
+ apply TApp with (A:=A); auto.

(* TPCase *)
+ apply TPCase with (t':=t') (A1:=A1) (A2:=A2); auto.

(* TEq *)   
+ apply TEq with (A:=A) (t1:=t1) (B:=B) (t2:=t2); auto.

(* TConv *)
+ apply TConv with (p:=p) (b1:=b1) (B1:=B1) (B2:=B2); auto.

(* TSubtype *)
+ apply TSubtype with (A:=A); auto.

(* TSubtype_L *)
+ apply TSubtype_L with (A:=A); auto.

(* CWFNil *)
+ inversion H; subst; auto.

(* CWFTrm *)
+ inversion H1; subst.
   apply CWFTrm; auto.

   destruct y as [t' A'].
   apply CWFTrm; auto.
   destruct (Hstronger t A t' A' H4) as [Hstronger_term Hstronger_type].
   replace A' with (lift A' 0 0) by  (apply lift_0_is_id).
   rewrite <- Hstronger_type.
   replace A with (lift A 0 0) in H by (apply lift_0_is_id).
   auto.
Qed.


(* The de bruijn operations meet FO *)
Theorem lift_preserves_fo : forall A,
  FO A -> forall n m, FO (lift A n m).
Proof.
  induction 1; simpl; intros; auto. 
Qed.
Hint Resolve lift_preserves_fo.

Theorem open_preserves_fo : forall A,
  FO A -> forall a n, FO ({n ~> a} A).
Proof.
  induction 1; simpl; intros; auto.
Qed.
Hint Resolve open_preserves_fo.

Theorem theta_strengthening_fo_typing_cons : forall B G t a A,
     FO B
  -> typing ((th_log, B) :: G) t a A
  -> typing ((th_prog, B) :: G) t a A.
Proof.
  intros B G t a A Hfo Htyping.
  apply strengthening_elementwise
     with (P := fun x y =>
                  match x, y with 
                      (th, A), (th', A') => A=A' /\ FO A
                      end)
          (G := (th_log,B) :: G).
   + clear.
     intros.
     destruct_things.
     split; intros; subst; try tauto; eauto.
   + auto. 
   + auto.
Qed.

Theorem theta_strengthening_fo_typing_cons_cons : forall th C B G t a A,
     FO B
  -> typing ((th,C) :: (th_log, B) :: G) t a A
  -> typing ((th,C) :: (th_prog, B) :: G) t a A.
Proof.
  intros.
  apply strengthening_elementwise
     with (P := fun x y =>
                  match x, y with 
                      (th, A), (th', A') => A=A' /\ FO A
                      end)
          (G := (th,C) :: (th_log,B) :: G).
   + clear.
     intros.
     destruct_things.
     split; intros; subst; try tauto; eauto.
   + auto. 
   + auto.
Qed.

Theorem theta_strengthening_fo_typing_cons_cons_cons : forall C th B1 B2 th' G t a A,
     FO B1 
  -> FO B2
  -> typing ((th,C) :: (th_log, B1) :: (th_log, B2) :: G) t a A
  -> typing ((th,C) :: (th', B1) :: (th', B2) :: G) t a A.
Proof.
  intros.
  apply strengthening_elementwise
     with (P := fun x y =>
                  match x, y with 
                      (th, A), (th', A') => A=A' /\ FO A
                      end)
          (G := (th,C) :: (th_log,B1) :: (th_log,B2) :: G).
   + clear.
     intros.
     destruct_things.
     split; intros; subst; try tauto; eauto.
   + auto. 
   + auto.
Qed.

(* Move this somewhere else *)
Lemma TAtLV' : forall (G:ctx) (v A:exp) th,
       val v
    -> typing G th v A
    -> typing G th_log A exp_star
    -> typing G th_log v (exp_at A th).
Proof.
  destruct th; auto.
Qed.

(* Move this somewhere else *)
Lemma TAtSame : forall (G:ctx) (t:theta) (a A:exp),
       typing G t a A
    -> typing G th_log A exp_star
    -> typing G t a (exp_at A t).
Proof.
  destruct t; auto.
Qed.

Lemma boxed_hyp : 
  forall B G t a A th0, 
       typing ((th_log, exp_at B th0) :: G) t a A
    -> typing ((th0, B) :: G) t a A.
Proof.
  intros.
  apply strengthening_elementwise
     with (P := fun x y =>
                  match x, y with 
                      (th, A), (th', A') => 
                        th=th_log /\ th'=th0 /\ A=exp_at A' th0
                      end)
          (G := ((th_log, exp_at B th0) :: G)).
  + clear.
    intros.
    destruct_things.
    subst.
    split.
      - simpl. intros.
        apply typ_at_inversion in H0; destruct H0. 
        apply TAtLV'; auto.
      - simpl. intros.
        split.
          intros; apply typ_at_inversion in H; destruct H; auto.
          auto.
  + assumption.   
  + auto.
Qed.


Lemma better_typ_refl' : forall G A Aty,
  typing G th_log A Aty 
  -> Aty = exp_star
  -> better_typ A A.
Proof.
  induction 1; intros Heq; try rewrite Heq in *;
  try congruence;
  try match goal with 
  | [ H : better_typ _ exp_star |- _]
      => elimtype False; now eauto using better_not_star
  | [ H : better_typ_L _ exp_star |- _]
      => elimtype False; now eauto using not_better_L_star
  | [ H : typing _ _ exp_star _ |- _ ]
      => elimtype False; now eauto using star_not_typable
  end;
  auto.
Qed.

Lemma better_typ_refl : forall G A,
  typing G th_log A exp_star 
  -> better_typ A A.
Proof.
  intros.
  eauto using better_typ_refl'.  
Qed.

Ltac kinding_inversion :=
  match goal with
   | [ H: typing _ _ (exp_at _ _)  exp_star |- _ ]
     => apply typ_at_inversion in H;  destruct_things
   | [ H: typing _ _ (exp_arr _ _) exp_star |- _ ]
     => apply typ_arr_inversion in H; destruct_things
  end.   

(* This weird-looking thing is what we need to justify the translation of nrec. *)
Lemma nrec_hyp :
 forall G B a A,
   typing 
     ((th_log,exp_at
		   (exp_arr exp_nat
		      (exp_at
			 (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
			    (lift (lift (translate_ttyp B) 1 1) 1 0))
			 th_log)) 
                   th_log)  :: G)
     th_log a A
  ->
  typing
     ((th_log,
      exp_arr exp_nat
        (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
           (lift (lift (translate_ttyp B) 1 1) 1 0)))
      :: G)
     th_log a A.
Proof.
  intros.
  apply strengthening_elementwise
   with (P := 
      fun x y =>
        match x, y with 
          | (th_log, exp_at (exp_arr 
                             exp_nat
                             (exp_at 
                                (exp_arr (exp_eq (exp_succ v1) v0) B)
                                th_log))
                             th_log),
            (th_log, exp_arr exp_nat 
                             (exp_arr (exp_eq (exp_succ v1') v0') B'))
            => v1=v1' /\ v0=v0' /\ B=B'
          | _,_ => False            
                  end)
      (G := ((th_log,
         exp_at
           (exp_arr exp_nat
              (exp_at
                 (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
                    (lift (lift (translate_ttyp B) 1 1) 1 0)) th_log)) th_log))
         :: G).
  + clear.
    intros th A th' A'.
    refine (match th, A, th', A' with 
             | th_log, 
               (exp_at (exp_arr 
                          exp_nat
                          (exp_at 
                             (exp_arr (exp_eq (exp_succ v1) v0) B)
                             th_log))
                       th_log),
               th_log,
               (exp_arr exp_nat 
                        (exp_arr (exp_eq (exp_succ v1') v0') B'))
               => _
             | _, _, _, _ => _            
           end); simpl; try tauto.
    intros.
    destruct_things.
    subst.
    split.
      - simpl. intros.
        apply TAtL.

        apply TSubtype_L with (A := (exp_arr exp_nat
            (exp_arr
               (exp_eq (exp_succ (lift v1' n (S m))) (lift v0' n (S m)))
               (lift B' n (S (S m)))))).
         assumption. 
         repeat kinding_inversion.
         apply BeTLArr.
           eauto using better_typ_refl.
           apply BeTLAtL.
             eauto using better_typ_refl.

         repeat kinding_inversion.
         auto.
         
         repeat kinding_inversion.
         auto.

      - simpl. intros.
        split; intros; repeat kinding_inversion; auto.
  + assumption.   
  + auto.
Qed.

Theorem translate_better_ttyp :
 forall A A', 
   better_ttyp A A'
   -> better_typ (translate_ttyp A) (translate_ttyp A').
Proof.
  Hint Constructors better_typ.
  induction 1; simpl; auto.
Qed.

Lemma translate_tval : forall v,
  tval v -> val (translate_texp v).
Proof.
  induction 1; simpl in *; auto.
Qed.

Lemma translate_ptstep : forall a b,
  ptstep a b -> pstep (translate_texp a) (translate_texp b).
Proof.
  induction 1; simpl in *; auto;
    repeat rewrite translate_open_texp in *; simpl;
    try rewrite translate_lift_texp; simpl; 
    auto using translate_tval.
Qed.
   
Lemma translate_mptstep : forall a b,
  mptstep a b -> mpstep (translate_texp a) (translate_texp b).
Proof.
  induction 1; eauto using translate_ptstep.
Qed.

Theorem translate_ttyping : forall G th a A,
  ttyping G th a A ->
  typing (translate_tctx G) th (translate_texp a) (translate_ttyp A).
Proof.
  apply (ttyping_ind_mut 
         (fun G th a A _ =>
            typing (translate_tctx G) th (translate_texp a) (translate_ttyp A))
         (fun G A _ =>
            typing (translate_tctx G) th_log (translate_ttyp A) exp_star)
         (fun G _ => ctx_wf (translate_tctx G)));
  intros; simpl in *; auto.

  (* TTVar *)  
+ rewrite translate_tctx_app in *.
  simpl in *.
  rewrite translate_lift_ttyp in *.
  apply TSub'.
  apply TVar.
   auto.
   rewrite translate_tctx_length; auto.
   auto.

  (* TTApp. *)
+ rewrite translate_open_ttyp.  
  apply TApp with (A := translate_ttyp A).
    apply typing_unbox with (th':=th2); auto.
    auto.
    rewrite <- translate_open_ttyp; eauto.

  (* TTLam *)
+ (* this case will be better if we can make @-inversion part of go_go_regularity. *)
  apply typ_at_inversion in H0; destruct H0.
  go_go_regularity.
  apply typing_boxV; destruct th2; auto using theta_strengthening_fo_typing_cons.

  (* TTRec *)
+ (* this case will be better if we can make @-inversion part of go_go_regularity. *)  
  apply typ_at_inversion in H0; destruct H0.
  go_go_regularity.
  apply typing_boxV.
  apply TRec. simpl.
    apply boxed_hyp.
    apply theta_strengthening_fo_typing_cons_cons. 
      auto.
      repeat rewrite translate_lift_ttyp in *.
        apply H.
    auto.
    auto.
    auto.

  (* TTNRec *)
+  apply typ_at_inversion in H0; destruct H0. 
   apply typing_boxV.
   apply TNRec.
     repeat rewrite translate_lift_ttyp in H.
     apply nrec_hyp.
       auto.
     auto.
     auto.
     auto.
    
  (* TTPair *)
+ rewrite translate_open_ttyp in *.
  apply TPair; auto. 

  (* TTPCase *)
+ go_go_regularity.
  apply TPCase with (t' := t) (A1 := translate_ttyp A1) (A2:=translate_ttyp A2).
  apply TAtSame.
    auto.
    auto.
    repeat rewrite translate_lift_texp in H0.
    repeat rewrite translate_lift_ttyp in H0.
    apply theta_strengthening_fo_typing_cons_cons_cons; now auto using ttyp_FO.
    auto.

  (* TTNCase *)
+ apply TNCase.
   auto.
   repeat rewrite translate_lift_ttyp in *.
   apply H0.
     repeat rewrite translate_lift_texp.
     rewrite <- translate_lift_ttyp.
     rewrite <- translate_lift_texp.
     destruct t.
       apply H1.
       apply theta_strengthening_fo_typing_cons_cons; now auto.
     auto.

  (* TTJoin *)
+ apply TSub'.
  eapply TJoin;
    try eapply translate_mptstep; eauto.

  (* TTConv *)
+ rewrite translate_open_ttyp in *.
  apply TConv with (p  := translate_texp p) 
                   (b1 := translate_texp b1)
                   (B1 := translate_ttyp B1)
                   (B2 := translate_ttyp B2); auto.

  (* TTSubtype *)
+ eauto using translate_better_ttyp.

  (* TKArr *)
+ auto using ttyp_FO, theta_strengthening_fo_typing_cons.

  (* TKSigma *)
+ auto using ttyp_FO, theta_strengthening_fo_typing_cons.

  (* TKEq *)
+ apply TEq with (t1:=t1) (t2:=t2) (A := translate_ttyp A) (B := translate_ttyp B);
    eauto. 
Qed.

Print Assumptions translate_ttyping.
