Require Import Arith.
(* Changes from the version in the paper:
    - Dropped TCtxt rule
    - Sigma type eliminator is pcase, not proj1/2.
    - Added term annotation to terminates,
      and type annotation to Terminates.
    - separate texp_fun and texp_nrec
    - Add kinding premise to application rule
  *)

Require Import Bool.
Require Import List.

Require Import shared_defs.

Inductive texp : Set := 
 | texp_var    (n:nat)
 | texp_fun    (a:texp)
 | texp_lam    (a:texp)
 | texp_nrec   (a:texp)
 | texp_app    (a:texp) (b:texp)
 | texp_zero   
 | texp_succ   (a:texp)
 | texp_ncase  (a:texp) (b1:texp) (b2:texp)
 | texp_pair   (a:texp) (b:texp)
 | texp_pcase  (a:texp) (b:texp)
 | texp_join
(*
 | texp_terminates
*)
 .

Inductive ttyp : Set := 
 | ttyp_nat 
 | ttyp_arr   (th:theta) (A:ttyp) (B:ttyp)
 | ttyp_eq    (a:texp) (b:texp)
 | ttyp_sigma (A:ttyp) (B:ttyp)
(*
 | ttyp_terminates (a:texp) (A:ttyp)
*)
 .

Definition tctx : Set := list ttyp.

(** subrules *)
Inductive tval : texp -> Prop :=
  | tval_var  : forall n, tval (texp_var n)
  | tval_fun  : forall a, tval (texp_fun a)
  | tval_lam  : forall a, tval (texp_lam a)
  | tval_nrec : forall a, tval (texp_nrec a)
  | tval_zero : tval texp_zero
  | tval_succ : forall v, tval v -> tval (texp_succ v)
  | tval_pair : forall v u, tval v -> tval u -> tval (texp_pair v u)
  | tval_join : tval texp_join
(*
  | tval_terminates : tval texp_terminates
 *)
  .

(** definitions *)
(* lift e n m  -  adds n to every var which is at least as big as m.

   so if     G1 ++ G3 |- e : A
   and       length G2 = n      length G1 = m
   then 
        (lift_c G1 n m) ++ G2 ++ G3 |- lift e n m : (lift A n m)
 *)
Fixpoint lift_texp (e : texp) (n m : nat) : texp :=
  match e with
  | texp_var n'        => texp_var (match nat_compare n' m with
                                  | Lt => n'
                                  | _  => (n' + n)
                                  end)
  | texp_fun  a        => texp_fun  (lift_texp a n (S (S m)))
  | texp_lam  a        => texp_lam  (lift_texp a n (S m))
  | texp_nrec a        => texp_nrec (lift_texp a n (S (S m)))
  | texp_app  a b      => texp_app  (lift_texp a n m) (lift_texp b n m)
  | texp_zero          => texp_zero
  | texp_succ a        => texp_succ (lift_texp a n m)
  | texp_ncase a b1 b2 => texp_ncase (lift_texp a n m) (lift_texp b1 n (S m))
                                   (lift_texp b2 n (S (S m)))
  | texp_pair a b      => texp_pair  (lift_texp a n m) (lift_texp b n m)
  | texp_pcase a b     => texp_pcase (lift_texp a n m) (lift_texp b n (S (S (S m))))
  | texp_join          => texp_join
(*
  | texp_terminates    => texp_terminates
 *)
  end.

Fixpoint lift_ttyp (t : ttyp) (n m : nat) : ttyp := 
  match t with
  | ttyp_nat            => ttyp_nat
  | ttyp_arr th A1 A2   => ttyp_arr th (lift_ttyp A1 n m) (lift_ttyp A2 n (S m))
  | ttyp_sigma A B      => ttyp_sigma (lift_ttyp A n m) (lift_ttyp B n (S m))
  | ttyp_eq a b         => ttyp_eq (lift_texp a n m) (lift_texp b n m)
 (*
  | ttyp_terminates a A => ttyp_terminates (lift_texp a n m) (lift_ttyp A n m) 
 *)
  end.

(* lift_c G1 n m - adds n to the term variables of G1 that refer at least
   m places outside of G1

   so if     |- G1 ++ G3
   and       length G2 = n
   then 
         |- (lift_c G1 n) ++ G2 ++ G3
 *)
Fixpoint lift_c (G : tctx) n m : tctx := 
  match G with
  | nil         => nil
  | A :: G' => lift_ttyp A n (m + length G') :: lift_c G' n m
  end.


Reserved Notation "{ n ~t> a }" (at level 60, no associativity).
Reserved Notation "{ n ~T> a }" (at level 60, no associativity).

Fixpoint open_texp (n : nat) (u : texp) (e : texp) {struct e} : texp :=
  match e with
  | texp_var  n'   => 
    match nat_compare n' n with
      | Lt => texp_var n'
      | Eq => u
      | Gt => texp_var (pred n')
    end
  | texp_fun  a    => texp_fun ({S (S n) ~t> lift_texp u 2 0} a)
  | texp_lam  a    => texp_lam ({S n ~t> lift_texp u 1 0} a)
  | texp_nrec  a   => texp_nrec ({S (S n) ~t> lift_texp u 2 0} a)
  | texp_app  a b  => texp_app ({n ~t> u}a) ({n ~t> u}b)
  | texp_zero      => texp_zero
  | texp_succ a    => texp_succ ({n ~t> u} a)
  | texp_ncase a b1 b2 => texp_ncase ({n ~t> u} a) ({S n ~t> lift_texp u 1 0}b1) 
                                   ({S (S n) ~t> lift_texp u 2 0} b2)
  | texp_pair a b  => texp_pair ({n ~t> u} a) ({ n ~t> u}b)
  | texp_pcase a b => texp_pcase ({n ~t> u} a) ({(S (S (S n))) ~t> lift_texp u 3 0}b)
  | texp_join      => texp_join
(*
  | texp_terminates => texp_terminates
 *)
  end
  where "{ n ~t> a }" := (open_texp n a).

Fixpoint open_ttyp (n : nat) (u : texp) (t : ttyp) {struct t} : ttyp :=
  match t with 
  | ttyp_nat            => ttyp_nat
  | ttyp_arr th A B     => ttyp_arr th ({n ~T> u}A) ({S n ~T> lift_texp u 1 0} B)
  | ttyp_sigma A B      => ttyp_sigma ({n ~T> u} A) ({S n ~T> lift_texp u 1 0}B)
  | ttyp_eq a b         => ttyp_eq ({n ~t> u}a) ({n ~t> u}b)
(*
  | ttyp_terminates a A => ttyp_terminates ({n ~t> u}a) ({n ~T> u}A)
*)
  end
  where "{ n ~T> a }" := (open_ttyp n a).

Fixpoint open_c (n : nat) (u : texp) (G : tctx) : tctx := 
  match G with
  | nil     => nil
  | A :: G' =>
       ({n + length G' ~T> lift_texp u (n + length G') 0} A) :: (open_c n u G')
  end.

Notation "{ n ~G> a }" := (open_c n a) 
  (at level 60, no associativity).


(* count the number of free variables *)
Fixpoint fvs_texp (a : texp) : nat :=
  match a with
  | texp_var n   => S n
  | texp_fun a   => pred (pred (fvs_texp a))
  | texp_lam a   => pred (fvs_texp a)
  | texp_nrec a  => pred (pred (fvs_texp a))
  | texp_app a b => max (fvs_texp a) (fvs_texp b)
  | texp_zero    => 0
  | texp_succ a  => fvs_texp a
  | texp_ncase a b1 b2 => max (fvs_texp a) (max (pred (fvs_texp b1))
                                          (pred (pred (fvs_texp b2))))
  | texp_pair a b  => max (fvs_texp a) (fvs_texp b)
  | texp_pcase a b => max (fvs_texp a) (pred (pred (pred (fvs_texp b))))
  | texp_join      => 0
(*
  | texp_terminates  => 0
*)
  end.

Fixpoint fvs_ttyp (t : ttyp) : nat :=
  match t with
  | ttyp_arr th A B     => max (fvs_ttyp A) (pred (fvs_ttyp B))
  | ttyp_nat            => 0
  | ttyp_sigma A B      => max (fvs_ttyp A) (pred (fvs_ttyp B))
  | ttyp_eq a b         => max (fvs_texp a) (fvs_texp b)
(*
  | ttyp_terminates a A => max (fvs_texp a) (fvs_ttyp A)
*)
  end.
 
(* defns JStepCBV *)
Inductive tstep : texp -> texp -> Prop :=    (* defn step *)
 | TSCaseZ : forall (a1 a2:texp),
     tstep (texp_ncase texp_zero a1 a2) ({0 ~t> texp_join} a1)
 | TSCaseS : forall (v a1 a2:texp),
     tval v ->
     tstep (texp_ncase (texp_succ v) a1 a2)  
          ({0 ~t> v} ({0 ~t> texp_join} a2))
 | TSCasePair : forall v1 v2 b,
     tval v1 ->
     tval v2 ->
     tstep (texp_pcase (texp_pair v1 v2) b)
          ({0 ~t> v1} ({0 ~t> lift_texp v2 1 0} ({0~t> texp_join}b)))
 | TSBetaFun : forall (a v:texp),
     tval v ->
     tstep (texp_app (texp_fun a) v)
          ({0 ~t> v} ({0 ~t> texp_fun (lift_texp a 1 2)} a))
 | TSBetaLam : forall (a v:texp),
     tval v ->
     tstep (texp_app (texp_lam a) v) ({0 ~t> v} a) 
 | TSBetaNRec : forall (a v:texp),
     tval v ->
     tstep (texp_app (texp_nrec a) v)
          ({0 ~t> v} 
           ({0 ~t> texp_lam (texp_lam (texp_app 
                     (texp_nrec (lift_texp a 3 2)) (texp_var 1)))} a))
 | TSSucc1 : forall a a',
     tstep a a' ->
     tstep (texp_succ a) (texp_succ a')
 | TSApp1 : forall (a a' b : texp),
     tstep a a' -> 
     tstep (texp_app a b) (texp_app a' b)
 | TSApp2 : forall (v b b' : texp),
     tval v -> 
     tstep b b' -> 
     tstep (texp_app v b) (texp_app v b')
 | TSNCase1 : forall (a a' b1 b2 : texp),
     tstep a a' -> 
     tstep (texp_ncase a b1 b2) (texp_ncase a' b1 b2)
 | TSPair1 : forall a a' b,
     tstep a a' ->
     tstep (texp_pair a b) (texp_pair a' b)
 | TSPair2 : forall v b b',
     tval v ->
     tstep b b' ->
     tstep (texp_pair v b) (texp_pair v b')
 | TSPCase1 : forall a a' b,
     tstep a a' ->
     tstep (texp_pcase a b) (texp_pcase a' b).


(** definitions *)
(* defns JMultiStep *)
Inductive mtstep : texp -> texp -> Prop :=    (* defn mstep *)
 | MTSRefl : forall (a:texp),
     mtstep a a
 | MSTStep : forall (a:texp) (b' b:texp),
     tstep a b -> 
     mtstep b b' -> 
     mtstep a b'.


(*******************************************************************)

Inductive ptstep : texp -> texp -> Prop :=
 | TPRefl : forall a, ptstep a a
 | TPCaseZ : forall (a1 a1' a2:texp),
      ptstep a1 a1'
   -> ptstep (texp_ncase texp_zero a1 a2) ({0 ~t> texp_join} a1')
 | TPCaseS : forall (v v' a1 a2 a2':texp),
      tval v
   -> ptstep v v'
   -> ptstep a2 a2'
   -> ptstep (texp_ncase (texp_succ v) a1 a2)
            ({0 ~t> v'} ({0 ~t> texp_join} a2'))
 | TPCasePair : forall  v1 v1' v2 v2' b b',
      tval v1 
   -> ptstep v1 v1'
   -> tval v2
   -> ptstep v2 v2'
   -> ptstep b b'
   -> ptstep (texp_pcase (texp_pair v1 v2) b)
            ({0 ~t> v1'} ({0 ~t> lift_texp v2' 1 0} ({0 ~t> texp_join} b')))
 | TPBetaFun : forall (a a' v v':texp),
      tval v 
   -> ptstep v v'
   -> ptstep a a'
   -> ptstep (texp_app (texp_fun a) v)
            ({0 ~t> v'} ({0 ~t> texp_fun (lift_texp a' 1 2)} a'))
 | TPBetaLam : forall (a a' v v':texp),
      tval v 
   -> ptstep v v'
   -> ptstep a a'
   -> ptstep (texp_app (texp_lam a) v) ({0 ~t> v'} a')
 | TPBetaNRec : forall (a a' v v':texp),
      tval v 
   -> ptstep v v'
   -> ptstep a a'
   -> ptstep (texp_app (texp_nrec a) v)
            ({0 ~t> v'}
             ({0 ~t> texp_lam (texp_lam (texp_app
                      (texp_nrec (lift_texp a' 3 2)) (texp_var 1)))} a'))
 | TPFun1 : forall (a a':texp),
      ptstep a a'
   -> ptstep (texp_fun a) (texp_fun a')
 | TPLam1 : forall (a a':texp),
      ptstep a a'
   -> ptstep (texp_lam a) (texp_lam a')
 | TPNRec1 : forall (a a':texp),
      ptstep a a'
   -> ptstep (texp_nrec a) (texp_nrec a')
 | TPSucc1 : forall a a',
      ptstep a a'
   -> ptstep (texp_succ a) (texp_succ a')
 | TPApp1: forall (a a' b b' : texp),
      ptstep a a'
   -> ptstep b b'
   -> ptstep (texp_app a b) (texp_app a' b')
 | TPNCase1 : forall (a a' b1 b1' b2 b2' : texp),
      ptstep a a'
   -> ptstep b1 b1'
   -> ptstep b2 b2'
   -> ptstep (texp_ncase a b1 b2) (texp_ncase a' b1' b2')
 | TPPCase1 : forall a a' b b',
      ptstep a a'
   -> ptstep b b'
   -> ptstep (texp_pcase a b) (texp_pcase a' b')
 | TPPair1 : forall (a a' b b' : texp),
      ptstep a a'
   -> ptstep b b'
   -> ptstep (texp_pair a b) (texp_pair a' b').


Inductive mptstep : texp -> texp -> Prop :=
  | MPTRefl : forall a, mptstep a a
  | MPTStep : forall a b c, ptstep a b -> mptstep b c -> mptstep a c.


(* Subtyping *)
Inductive better_ttyp : ttyp -> ttyp -> Prop :=
  | BTArr : forall th th' A A' B B',
    better_theta th th' 
    -> better_ttyp A' A
    -> better_ttyp B B'
    -> better_ttyp (ttyp_arr th A B) (ttyp_arr th' A' B')
  | BTSigma : forall A B B',
    better_ttyp B B'
    -> better_ttyp (ttyp_sigma A B) (ttyp_sigma A B')
  | BTNat : better_ttyp ttyp_nat ttyp_nat
  | BTEq : forall a b, better_ttyp (ttyp_eq a b) (ttyp_eq a b)
(*
  | BTTerminates : forall a A, better_ttyp (ttyp_terminates a A) (ttyp_terminates a A)
*)
  .


Inductive tctx_wf : tctx -> Prop :=
 | TCWFNil  : tctx_wf nil
 | TCWFTrm : forall A G,
       tkinding G A 
    -> tctx_wf G
    -> tctx_wf (A :: G)

with ttyping : tctx -> theta -> texp -> ttyp -> Prop :=
 | TTVar : forall (G1 G3:tctx) (th:theta) (n:nat) (A:ttyp),
     tctx_wf (G1 ++ A :: G3) ->
     tkinding (G1 ++ A :: G3) (lift_ttyp A (S n) 0) ->
     n = length G1 ->
     ttyping (G1 ++ A :: G3) th (texp_var n) (lift_ttyp A (S n) 0)
 | TTApp : forall (G:tctx) (th1 th2:theta) (a b : texp) (A B:ttyp),
       ttyping G th1 a (ttyp_arr th2 A B)
    -> ttyping G th1 b A
    -> better_theta th2 th1
    -> tkinding G  ({0 ~T> b} B)
    -> ttyping G th1 (texp_app a b) ({0 ~T> b} B)
 | TTLam : forall (G:tctx) th1 th2 (b :texp) (A B:ttyp),
       ttyping (A :: G) th2 b B
    -> tkinding G (ttyp_arr th2 A B)
    -> ttyping G th1 (texp_lam b) (ttyp_arr th2 A B)
 | TTRec : forall (G:tctx) th (a :texp) (A B:ttyp),
     ttyping ((lift_ttyp (ttyp_arr th_prog A B) 1 0) :: A :: G)
            th_prog a (lift_ttyp B 1 0) ->
     tkinding G (ttyp_arr th_prog A B) ->
     ttyping G th (texp_fun a) (ttyp_arr th_prog A B)
 | TTNRec : forall (G:tctx) th (a :texp)  (B:ttyp),
     ttyping (   (ttyp_arr th_log ttyp_nat 
                     (ttyp_arr th_log (ttyp_eq (texp_succ (texp_var 0))
                                               (texp_var 1))
                              (lift_ttyp (lift_ttyp B 1 1) 1 0)))
             :: ttyp_nat :: G)
            th_log a (lift_ttyp B 1 0) ->
     tkinding G (ttyp_arr th_log ttyp_nat B) ->
     ttyping  G th (texp_nrec a) (ttyp_arr th_log ttyp_nat B)
 | TTZero : forall (G:tctx) (th:theta),
     tctx_wf G ->
     ttyping G th texp_zero ttyp_nat
 | TTSuc : forall (G:tctx) (th:theta) (b:texp),
     ttyping G th b ttyp_nat ->
     ttyping G th (texp_succ b) ttyp_nat
 | TTPair : forall G th A B a b,
      tkinding G (ttyp_sigma A B)
   -> ttyping  G th a A
   -> tkinding G ({0 ~T> a}B)
   -> ttyping  G th b ({0~T>a}B)
   -> ttyping  G th (texp_pair a b) (ttyp_sigma A B)

 | TTPCase : forall G t A1 A2 B a b,
      ttyping G t a (ttyp_sigma A1 A2)
   -> ttyping (ttyp_eq (texp_pair (texp_var 1) (texp_var 0)) (lift_texp a 2 0)
              :: A2 
              :: A1
              :: G)
             t b (lift_ttyp B 3 0)
   -> tkinding G B
   -> ttyping G t (texp_pcase a b) B

 | TTNCase : forall (G:tctx) (t:theta) (a b1 b2 :texp) (A:ttyp),
     ttyping G t a ttyp_nat ->
     ttyping ((ttyp_eq texp_zero a) :: G) t b1 (lift_ttyp A 1 0) ->
     ttyping ((ttyp_eq (texp_succ (texp_var 0)) (lift_texp a 1 0)) 
               :: ttyp_nat :: G)
             t b2 (lift_ttyp A 2 0) ->
     tkinding G A ->
     ttyping G t (texp_ncase a b1 b2) A 
 | TTJoin : forall G a b c t1 t2 th A B,
       mptstep a c -> mptstep b c
    -> ttyping G t1 a A -> ttyping G t2 b B
    -> tkinding G A -> tkinding G B
    -> ttyping G th texp_join (ttyp_eq a b)
 | TTConv : forall G p a b1 b2 B1 B2 t A,
       ttyping G th_log p (ttyp_eq b1 b2)
    (* These two lines follow by regularity from the previous one,
       but improves the induction (ideally we should define an equivalent "clean" system): *)
    -> ttyping G th_prog b1 B1 -> tkinding G B1
    -> ttyping G th_prog b2 B2 -> tkinding G B2
    -> ttyping G t a ({0 ~T> b1}A)
    -> tkinding G ({0 ~T> b2}A)
    -> ttyping G t a ({0 ~T> b2}A)
(*
 | TTReify : forall G a A th,
       ttyping G th_log a A
    -> ttyping G th texp_terminates (ttyp_terminates a A)
 | TTReflect : forall G a A th p,
       ttyping G th_prog a A
    -> ttyping G th_log p (ttyp_terminates a A)
    -> ttyping G th a A
*)
 | TTSubtype : forall G th a A A',
       ttyping G th a A
    -> better_ttyp A A'
    -> tkinding G A'
    -> ttyping G th a A'

with tkinding : tctx -> ttyp -> Prop :=
 | TKArr : forall G A th B,
       tkinding G A 
    -> tkinding (A :: G) B
    -> tkinding G (ttyp_arr th A B)
 | TKSigma : forall G A B,
       tkinding G A 
    -> tkinding (A :: G) B
    -> tkinding G (ttyp_sigma A B)
 | TKNat : forall G,
     tctx_wf G -> tkinding G ttyp_nat
 | TKEq  : forall G a b A t1 B t2,
       ttyping G t1 a A -> ttyping G t2 b B
    -> tkinding G A -> tkinding G B
    -> tkinding G (ttyp_eq a b)
(*
 | TKTerminates : forall G a A,
       ttyping G th_prog a A
    -> tkinding G A  (* This one follows by regularity, but I'll postpone proving that and add this premise instead. *)
    -> tkinding G (ttyp_terminates a A)
*)
 .

Scheme ttyping_ind_mut  := Induction for ttyping Sort Prop
  with tkinding_ind_mut := Induction for tkinding Sort Prop                        
  with tctx_wf_ind_mut  := Induction for tctx_wf Sort Prop.

(********************************************************************************)

(* End of definitions. Below this line is infrastructure for proving stuff,
   followed by some actual theorems (who may eventually get to live in a different file). *)

Hint Constructors tval.

Hint Constructors better_theta.
Hint Resolve BetterLog BetterProg.
Hint Constructors ttyping.
Hint Constructors tkinding.
Hint Constructors tctx_wf.

Hint Constructors better_ttyp.

Lemma better_ttyp_refl : forall A, 
  better_ttyp A A.
induction A; auto.
Qed.
Hint Resolve better_ttyp_refl.

Lemma better_than_log : forall th1 th2,
  th2 = th_log -> better_theta th1 th2 -> th1 = th_log.
inversion 2; congruence.
Qed.

Lemma subsume_theta' : forall G th a A,
 ttyping G th a A -> th = th_log -> ttyping G th_prog a A.
Proof.
induction 1; auto; eauto.
Qed.

(* This lemma is quoted in the thesis *)
Lemma subsume_theta : forall G th a A,
 ttyping G th a A -> ttyping G th_prog a A.
Proof.
  destruct th;
    eauto using subsume_theta'.
Qed.

(* This lemma is quoted in the thesis *)
Lemma value_any_theta : forall G th v A th',
    ttyping G th v A 
 -> tval v
 -> ttyping G th' v A.
Proof.
induction 1; intros Hval; auto; eauto;
 inversion Hval; auto.
Qed.
