Require Omega.
Require Import Arith. 
Require Import Max.

Require Import shared_defs.
Require Import shared_tactics.
Require Import teqt_defs.
Require Import teqt_db_basics.
Require Import teqt_reduction.
Require Import teqt_confluence.

Fixpoint size  (A : ttyp) : nat :=
  match A with
    | ttyp_nat => 1
    | ttyp_arr th A1 A2 => S (max (size A1) (size A2)) 
    | ttyp_eq a1 a2 => 1
    | ttyp_sigma A1 A2 => S (max (size A1) (size A2))
(*
    | ttyp_terminates a1 A1 => S (size A1) 
*)
  end.

Lemma size_open : forall A n a, size ({n~T>a}A) = size A.
Proof.
  induction A; intros; simpl; auto.
Qed.

Definition closed_tval a :=
  tval a /\ fvs_texp a = 0.

Definition joinable (a1 a2 : texp) : Prop :=
  exists c, mptstep a1 c /\ mptstep a2 c.


(* A term is a numeral if it is built from texp_zero and texp_succ. *)
Inductive tnumeral : texp -> Prop :=
  | TNZero : tnumeral texp_zero
  | TNSucc : forall a, tnumeral a -> tnumeral (texp_succ a).
Hint Constructors tnumeral.

Hint Constructors tval.
Lemma tnumeral_closed_tval : forall a,
 tnumeral a -> closed_tval a.
Proof.
 induction 1; unfold closed_tval in *; destruct_things; simpl;
    eauto.
Qed.
Hint Resolve tnumeral_closed_tval.

Fixpoint interpV' (s : nat) (A : ttyp)  {struct s} : texp -> Prop := 
  match s with
   | 0 => fun a => False
   | S s' =>
     match A with
       | ttyp_nat => tnumeral
       | ttyp_arr th_prog A1 A2 => closed_tval
       | ttyp_arr th_log A1 A2 =>
         fun a =>
          closed_tval a /\
               forall v, interpV' s' A1 v 
                 -> exists r, interpV' s' ({0 ~T> v} A2) r
                              /\ mtstep (texp_app a v) r
       | ttyp_eq a1 a2 => fun a => 
                          a = texp_join /\ joinable a1 a2
       | ttyp_sigma A1 A2 => fun a => 
                          closed_tval a /\
                          match a with
                            | texp_pair a1 a2 =>
                                interpV' s' A1 a1 /\ interpV' s' ({0 ~T> a1} A2) a2
                            | _ => False 
                          end
(*
       | ttyp_terminates a1 A1 => fun a => 
                          a = texp_terminates /\ exists b, tval b /\ mptstep a1 b
*)
     end
  end.


Lemma interpV'_n : forall n A a, size A <= n -> (interpV' (size A) A a <-> interpV' n A a).
induction n using lt_wf_ind.
  intros A a.
  destruct A;
    (destruct n; simpl; intros Hle; inversion Hle; subst).
    + tauto.
    + tauto.
    + destruct th; destruct a; intuition.
    + destruct th;  intuition.
      (*  -> implication: *)
      - assert (interpV' (max (size A1) (size A2)) A1 v) as interpV_v.
          rewrite <- H.
          rewrite -> H with (m:=n). 
          assumption.
          auto with arith. eauto with arith. eauto with arith. auto with arith.
        destruct (H3 v interpV_v) as [r [interpV_r steps]].
        exists r. split.
          rewrite <- H by (try rewrite size_open; eauto with arith).
          rewrite <- H in interpV_r by (try rewrite size_open; eauto with arith).
          assumption.
          assumption.
      (* <- implication: *)
      - assert (interpV' n A1 v) as interpV_v.
          rewrite <- H
            by eauto with arith.
          rewrite H with (m := (max (size A1) (size A2)))
            by eauto with arith.
          assumption.
        destruct (H3 v interpV_v) as [r [interpV_r steps]].
        exists r. split.
          rewrite <- H by (try rewrite size_open; eauto with arith).
          rewrite <- H in interpV_r by (try rewrite size_open; eauto with arith).
          assumption.
          assumption.
    + tauto.
    + tauto.
    + tauto.
    + destruct a; intuition.
      - rewrite <- H by eauto with arith.
        rewrite H with (m:= (max (size A1) (size A2))); eauto with arith.
      - rewrite <- H by (try rewrite size_open; eauto with arith).
        rewrite H with (m := (max (size A1) (size A2)));
         (try rewrite size_open; eauto with arith).
      - rewrite <- H by (eauto with arith).
        rewrite H with (m:=n); eauto with arith.
      - rewrite <- H by (try rewrite size_open; eauto with arith).
        rewrite H with (m:=n);
          (try rewrite size_open; eauto with arith).
Qed.

Definition interpV A := interpV' (size A) A.

Definition  interpC  (A:ttyp) (a:texp) := 
  exists r, interpV A r /\ mtstep a r.

Lemma interpV_nat : forall v,
  interpV ttyp_nat v <-> tnumeral v.
Proof.
  intros; unfold interpV; simpl; tauto.
Qed.

Lemma interpV_arr_log : forall A1 A2 a,
  interpV (ttyp_arr th_log A1 A2) a 
  <-> 
  (closed_tval a 
    /\ forall v, interpV A1 v ->  
             interpC ({0~T> v}A2) (texp_app a v)).
Proof.
  intros; unfold interpV; simpl.
  intuition.
    - rewrite interpV'_n  with (n := max (size A1) (size A2)) in H
                          by auto with arith.
      destruct (H1 v H) as [r [interpV_r steps]].
      exists r. split.
        unfold interpV.
        rewrite interpV'_n with (n := max (size A1) (size A2))
          by (try rewrite size_open; auto with arith).
        assumption.
        assumption.
    - rewrite <- interpV'_n in H
                          by auto with arith.
      destruct (H1 v H) as [r [interpV_r steps]].
      exists r. split.
        rewrite <- interpV'_n
          by (try rewrite size_open; auto with arith).
        assumption.
        assumption.
Qed.

Lemma interpV_arr_prog : forall A1 A2 a,
  interpV (ttyp_arr th_prog A1 A2) a 
  <->
  (closed_tval a).
Proof.
  intros; destruct a; unfold interpV; simpl; try tauto.
Qed.

Lemma interpV_sigma : forall A1 A2 a, 
  interpV (ttyp_sigma A1 A2) a
  <->
  (closed_tval a /\
   match a with
    | texp_pair a1 a2 => 
        interpV A1 a1 /\ interpV ({0 ~T> a1}A2) a2
    | _ => False
   end).
Proof.
  intros; destruct a; unfold interpV; simpl; intuition.
    + rewrite interpV'_n with (n := max (size A1) (size A2));
        eauto with arith.
    + rewrite interpV'_n with (n := max (size A1) (size A2));
        try rewrite size_open; eauto with arith.
    + rewrite <- interpV'_n;
        eauto with arith.
    + rewrite <- interpV'_n.
        assumption.
        try rewrite size_open; eauto with arith.
Qed.

Lemma interpV_eq : forall a1 a2 v, 
  interpV (ttyp_eq a1 a2) v
  <->
  (v = texp_join /\ joinable a1 a2).
Proof.
 intros; unfold interpV; simpl; tauto.
Qed.

Lemma closed_tval__tval : forall a, 
  closed_tval a -> tval a.
Proof.
  unfold closed_tval; tauto.
Qed.

Lemma closed_tval__closed : forall a,
  closed_tval a -> fvs_texp a = 0.
Proof.
  unfold closed_tval; tauto.
Qed.
  
Lemma interpV_val : forall A v, interpV A v -> tval v.
Proof.
  Hint Resolve closed_tval__tval.
  Hint Constructors tval.
  destruct A; try destruct th; destruct v; unfold interpV; simpl; 
    intros; destruct_things; try tauto; try congruence; auto.
Qed. 

Lemma interpV_closed : forall A v, interpV A v -> fvs_texp v = 0.
Proof.
  Hint Resolve closed_tval__closed.
  Hint Constructors tval.
  destruct A; try destruct th; intros v H;
  unfold interpV in H; simpl in H; 
    destruct_things; subst; simpl; try tauto; try congruence; auto.
Qed. 

Lemma better_ttyp_open : forall A B,
  better_ttyp A B
  -> forall  x v, better_ttyp ({x ~T> v}A) ({x ~T> v}B).
Proof.
 induction 1; intros; simpl; auto.
Qed.
          
Require Import List.

(*** Environments, i.e. multi-substitutions. ***)
Definition env := list texp.

Fixpoint close_texp (rho : env) (a : texp) (m:nat) : texp :=
  match rho with
   | nil => a
   | v :: rho' => close_texp rho' ({m ~t> v} a) m
  end.

Fixpoint close_ttyp (rho : env) (A : ttyp) (m:nat): ttyp :=
  match rho with
    | nil => A
    | v :: rho' => close_ttyp rho' ({m ~T> v} A) m
  end.

Inductive closed_env : env  -> Prop :=
  | ClNil  : closed_env nil
  | ClCons : forall (a : texp)
                    (H : fvs_texp a = 0)
                    (rho' : env)
                    (cl_rho' : closed_env rho'),
              closed_env (a::rho').
Hint Constructors closed_env.

Inductive wf_env : tctx -> env  -> Prop :=
  | ENil  : wf_env nil nil
  | ECons : forall G 
                   (A : ttyp)
                   (a : texp)
                   (rho' : env)
                   (wf_rho' : wf_env G rho')
                   (H_interpV : interpV (close_ttyp rho' A 0) a),
              wf_env (A :: G) (a::rho').
Hint Constructors wf_env.

Lemma wf_env_closed : forall G rho,
  wf_env G rho -> closed_env rho.
Proof.
  induction 1; eauto using interpV_closed.
Qed.

Lemma close_cong_var_gt :
 forall rho m n,
   m>n ->
   close_texp rho (texp_var n) m = (texp_var n).
Proof.
  induction rho; intros; simpl; auto.
  db_var_case.
Qed.

Lemma close_cong_nat :
  forall rho m,
    close_ttyp rho ttyp_nat m = ttyp_nat.
Proof.
  induction rho; simpl; intros; rewrite_ih; auto.
Qed.

Lemma close_cong_arr :
  forall rho,
    closed_env rho ->
    forall th A B m,
     close_ttyp rho (ttyp_arr th A B) m
      = ttyp_arr th (close_ttyp rho A m) (close_ttyp rho B (S m)).
Proof.
  induction 1; simpl; intros.
    + auto.    
    + rewrite IHclosed_env.
      rewrite lift_closed_texp_id by auto.
      auto.
Qed.

Lemma close_cong_sigma :
  forall rho,
    closed_env rho ->
    forall A B m,
     close_ttyp rho (ttyp_sigma A B) m
      = ttyp_sigma  (close_ttyp rho A m) (close_ttyp rho B (S m)).
Proof.
  induction 1; simpl; intros.
    + auto.    
    + rewrite IHclosed_env.
      rewrite lift_closed_texp_id by auto.
      auto.
Qed.

Lemma close_cong_eq : 
  forall rho m,
    forall a b,
      close_ttyp rho (ttyp_eq a b) m
       = ttyp_eq (close_texp rho a m) (close_texp rho b m).
Proof.
  induction rho; simpl; intros; auto.
Qed.

Lemma close_cong_fun :
  forall rho,
    closed_env rho ->
    forall a m,
     close_texp rho (texp_fun a) m
      = texp_fun  (close_texp rho a (S (S m))).
Proof.
  induction 1; simpl; intros.
    + auto.    
    + rewrite IHclosed_env.
      rewrite lift_closed_texp_id by auto.
      auto.
Qed.

Lemma close_cong_nrec :
  forall rho,
    closed_env rho ->
    forall a m,
     close_texp rho (texp_nrec a) m
      = texp_nrec  (close_texp rho a (S (S m))).
Proof.
  induction 1; simpl; intros.
    + auto.    
    + rewrite IHclosed_env.
      rewrite lift_closed_texp_id by auto.
      auto.
Qed.

Lemma close_cong_lam :
  forall rho,
    closed_env rho ->
    forall a m,
     close_texp rho (texp_lam a) m
      = texp_lam  (close_texp rho a (S m)).
Proof.
  induction 1; simpl; intros.
    + auto.    
    + rewrite IHclosed_env.
      rewrite lift_closed_texp_id by auto.
      auto.
Qed.

Lemma close_cong_app : 
  forall rho m,
    forall a b,
      close_texp rho (texp_app a b) m
       = texp_app (close_texp rho a m) (close_texp rho b m).
Proof.
  induction rho; simpl; intros; auto.
Qed.


Lemma close_cong_pair : 
  forall rho m,
    forall a b,
      close_texp rho (texp_pair a b) m
       = texp_pair (close_texp rho a m) (close_texp rho b m).
Proof.
  induction rho; simpl; intros; auto.
Qed.

Lemma close_cong_pcase : 
  forall rho,
    closed_env rho ->
    forall m a b,
      close_texp rho (texp_pcase a b) m
       = texp_pcase (close_texp rho a m) (close_texp rho b (S (S (S m)))).
Proof.
  induction 1; simpl; intros; auto.
  rewrite IHclosed_env.
  rewrite lift_closed_texp_id; auto.
Qed.  

Lemma close_cong_ncase : 
  forall rho,
    closed_env rho ->
    forall m a b1 b2,
      close_texp rho (texp_ncase a b1 b2) m
       = texp_ncase (close_texp rho a m) 
                    (close_texp rho b1 (S m))
                    (close_texp rho b2 (S (S m))).
Proof.
  induction 1; simpl; intros; auto.
  rewrite IHclosed_env.
  repeat rewrite lift_closed_texp_id; auto.
Qed.

Lemma close_cong_zero :
  forall rho m,
      close_texp rho texp_zero m
       = texp_zero.
Proof.
  induction rho; simpl; intros; auto.
Qed.

Lemma close_cong_succ : 
  forall rho m a,
      close_texp rho (texp_succ a) m
       = texp_succ (close_texp rho a m).
Proof.
  induction rho; simpl; intros; auto.
Qed.

Hint Rewrite close_cong_nat close_cong_arr close_cong_sigma close_cong_eq
   : step.

Theorem close_preserves_mptstep : 
  forall G rho,
    wf_env G rho ->
    forall a1 a2 n,
     mptstep a1 a2 ->
     mptstep (close_texp rho a1 n) (close_texp rho a2 n).
Proof.
  induction 1; simpl; auto.
  intros.
  apply IHwf_env.  
  apply open_preserves_mptstep; eauto using interpV_val.
Qed.   

Lemma close_closed_texp_id_gen : 
  forall (rho : env) a m,
    fvs_texp a <= m -> close_texp rho a m  = a.
Proof.
  induction rho; intros.
    auto.
    simpl. rewrite open_closed_texp_id_gen; auto.
Qed.

Lemma close_closed_texp_id : 
  forall (rho : env) a m,
    fvs_texp a = 0 -> close_texp rho a m  = a.
Proof.
  intros.
  apply close_closed_texp_id_gen; auto.
Qed.

Lemma open_close_texp_swap_gen : 
  forall rho, 
    closed_env rho ->
    forall b a i n,
    i <= n ->
    close_texp rho ({i ~t> b}a) n
    = ({i ~t> close_texp rho b n}) (close_texp rho a (S n)).
Proof.
  induction 1; intros; simpl.
    reflexivity.
   
    rewrite open_open_texp_swap by omega.
    rewrite IHclosed_env.
    rewrite lift_closed_texp_id by auto.
    reflexivity.
    auto.
Qed.

Lemma open_close_texp_swap : 
  forall rho, 
    closed_env rho ->
    forall b a n,
    close_texp rho ({0 ~t> b}a) n
    = ({0 ~t> close_texp rho b n}) (close_texp rho a (S n)).
Proof.
  induction 1; intros; simpl.
    reflexivity.
   
    rewrite open_open_texp_swap by omega.
    rewrite IHclosed_env.
    rewrite lift_closed_texp_id by auto.
    reflexivity.
Qed.

Lemma open_close_ttyp_swap : 
  forall rho, 
    closed_env rho ->
    forall b B n,
    close_ttyp rho ({0 ~T> b}B) n
    = ({0 ~T> close_texp rho b n}) (close_ttyp rho B (S n)).
Proof.
  induction 1; intros; simpl.
    reflexivity.
   
    rewrite open_open_ttyp_swap by omega.
    rewrite IHclosed_env.
    rewrite lift_closed_texp_id by auto.
    reflexivity.
Qed.

Lemma open_close_closed_texp_swap : 
  forall rho, 
    closed_env rho ->
    forall b a n,
      fvs_texp b = 0 ->
        close_texp rho ({0 ~t> b}a) n
         = ({0 ~t> b}) (close_texp rho a (S n)).
Proof.
  intros.
  rewrite open_close_texp_swap; auto.
  rewrite close_closed_texp_id; auto.
Qed.

Lemma open_close_closed_texp_swap_gen : 
  forall rho, 
    closed_env rho ->
    forall b a i n,
    fvs_texp b = 0 ->
    i <= n ->
    close_texp rho ({i ~t> b}a) n
    = ({i ~t> b}) (close_texp rho a (S n)).
Proof.
  intros.
  rewrite open_close_texp_swap_gen; auto.
  rewrite close_closed_texp_id; auto.
Qed.

Lemma open_close_closed_ttyp_swap : 
  forall rho, 
    closed_env rho ->
    forall b B n,
      fvs_texp b = 0 ->
        close_ttyp rho ({0 ~T> b}B) n
         = ({0 ~T> b}) (close_ttyp rho B (S n)).
Proof.
  intros.
  rewrite open_close_ttyp_swap; auto.
  rewrite close_closed_texp_id; auto.
Qed.

Lemma close_lift_ttyp_swap : forall rho ,
  closed_env rho -> forall A n k,
  close_ttyp rho (lift_ttyp A n 0) (n+k) 
  = (lift_ttyp (close_ttyp rho A k) n 0).
Proof.
  induction 1 ; simpl; intros; auto.
  replace a with (lift_texp a (n+k) 0) by (auto using lift_closed_texp_id).
  rewrite open_lift_ttyp_commute; try omega.
  rewrite IHclosed_env.
  replace (lift_texp a (n+k) 0) with a by (rewrite lift_closed_texp_id; auto).
  replace (lift_texp a k 0) with a by (rewrite lift_closed_texp_id; auto).
  auto.
Qed.


(* This comes up a lot: values which are known to be in interpV are
   closed, so closing them has no effect. *)
Hint Extern 3 =>
  match goal with
   | [ H1: interpV _ ?v,
       H2: wf_env ?G0 ?rho 
        |- close_texp ?rho ?v 0 = ?v 
     ] =>  rewrite close_closed_texp_id with (G:=G0);  now eauto using interpV_closed
   | [ H1: interpV _ ?v,
       H2: wf_env ?G0 ?rho 
        |- ?v = close_texp ?rho ?v 0
     ] =>  rewrite close_closed_texp_id with (G:=G0);  now eauto using interpV_closed
  end.

Lemma joinable_refl : forall a,
                        joinable a a.
Proof.
  intros; exists a; eauto.
Qed.
Hint Resolve joinable_refl.

Lemma joinable_symm : forall a1 a2,
  joinable a1 a2 -> joinable a2 a1.
Proof.
  intros a1 a2 [c [H1 H2]].
  unfold joinable.
  eauto.
Qed.
Hint Resolve joinable_symm.

Lemma mtstep_joinable : forall a1 a2,
 mtstep a1 a2 -> joinable a1 a2.
Proof.
  intros.
  apply mtstep_mptstep in H.
  exists a2; auto.  
Qed.
Hint Resolve mtstep_joinable.

Lemma open_joinable_into : forall a1 a2 b m,
  joinable a1 a2 -> joinable ({m ~t> a1}b) ({m ~t> a2}b).
Proof.
  intros a1 a2 b m [c [H1 H2]].
  exists ({m ~t> c}b);
    eauto using open_mptstep_into.
Qed.

Lemma MPTTrans : forall a b,
  mptstep a b -> forall c, mptstep b c -> mptstep a c.
Proof.
  induction 1.
   + auto.
   + intros c0.
     intros.
     apply IHmptstep in H1.
     apply MPTStep with (b:=b); auto.
Qed.

Lemma joinable_trans : 
  forall a b c,
      joinable a b 
   -> joinable b c
   -> joinable a c.
Proof.
  unfold joinable.
  intros a b c [c1 [H1 H1']] [c2 [H2 H2']].
  destruct (diamond_mpstep _ _ H1' _ H2) as [c3 [H3 H3']].
  exists c3; eauto using MPTTrans.
Qed.

Lemma joinable_joinable : forall b1 b2 a1 a2 m,
 joinable b1 b2 
 -> joinable ({m ~t> b1}a1) ({m ~t> b1}a2)
 -> joinable ({m ~t> b2}a1) ({m ~t> b2}a2).
Proof.
 intros.
 apply joinable_trans with ({m ~t> b1}a1); 
   auto using open_joinable_into, joinable_symm.
 apply joinable_trans with ({m ~t> b1}a2); 
   auto.
 auto using open_joinable_into.
Qed.

Inductive joinable_env : env -> env  -> Prop :=
  | JNil  : joinable_env nil nil
  | JCons : forall a a' rho rho',
                 joinable a a' 
              -> joinable_env rho rho'
              -> joinable_env (a::rho) (a'::rho').
Hint Constructors joinable_env.

Lemma joinable_env_cons : forall a rho rho',
  joinable_env rho rho'
  -> joinable_env (a::rho) (a::rho').
Proof.
  intros.
  apply JCons.
    exists a; auto.
    auto.
Qed.
Hint Resolve joinable_env_cons.

Lemma joinable_env_symm : forall rho rho',
  joinable_env rho rho' -> joinable_env rho' rho.
Proof.
  induction 1;
    eauto using joinable_symm.
Qed.
Hint Resolve joinable_env_symm.

Lemma joinable_env_open :
 forall rho rho',
    closed_env rho ->
    closed_env rho' ->
    joinable_env rho rho' ->
    forall a1 a2 n, 
      joinable (close_texp rho  a1 n) (close_texp rho  a2 n) ->
      joinable (close_texp rho' a1 n) (close_texp rho' a2 n).
Proof.
 intros rho rho' Hclosed Hclosed'.
 induction 1.
   + simpl; auto.
   + inversion Hclosed.
     inversion Hclosed'.
     subst.
     intros a1 a2 n H_joinable_rho.
     simpl in *.
     repeat rewrite open_close_closed_texp_swap_gen by (auto with arith).
     repeat rewrite open_close_closed_texp_swap_gen in H_joinable_rho by (auto with arith).
     apply joinable_joinable with a.
       assumption.

     repeat rewrite <- open_close_closed_texp_swap_gen by (auto with arith).
     repeat rewrite <- open_close_closed_texp_swap_gen in H_joinable_rho by (auto with arith).

     apply IHjoinable_env; auto.
Qed.

(* The only interesting case is for equals, all the other ones 
   are just trivially passing along the equivalence.
   There is probably a better way of structuring the proof... *)
Lemma interpV_respects_joinable : forall A rho rho' v,
  closed_env rho ->
  closed_env rho' ->
  joinable_env rho rho' ->
    interpV (close_ttyp rho A 0) v -> interpV (close_ttyp rho' A 0) v.
Proof.
  induction A; intros.
  (* nat *)
  + rewrite close_cong_nat in *.
    auto.
  (* arrow *)
  + rewrite close_cong_arr  by assumption.
    rewrite close_cong_arr  in H2 by assumption.
    destruct th.
    * (* th is log *)
      rewrite interpV_arr_log in *.
      unfold interpC in *.
      destruct_things.
      split; auto.
      intros v0 Hinterp.
      assert (interpV (close_ttyp rho A1 0) v0) as Hinterp'
         by (apply IHA1 with (rho:=rho'); auto).
      destruct (H3 v0 Hinterp') as [r [Hr_interp Hr_step]].
      clear H3.
      rewrite <- open_close_closed_ttyp_swap by (eauto using interpV_closed).
      change (close_ttyp rho' (({0 ~T> v0}) A2) 0)
        with (close_ttyp (v0::rho') A2 0).
      rewrite <- open_close_closed_ttyp_swap in Hr_interp by (eauto using interpV_closed).
      change (close_ttyp rho (({0 ~T> v0}) A2) 0)
        with (close_ttyp (v0::rho) A2 0) in Hr_interp.
      exists r. 
      split.
        apply IHA2 with (rho:=v0::rho);(eauto using interpV_closed, interpV_val). 
        auto.
    * (* th is prog *)
      rewrite interpV_arr_prog in *; auto.
  (* equals *)
  + rewrite close_cong_eq in *. 
    rewrite interpV_eq in *.
    intuition.
    apply joinable_env_open with (rho:=rho); auto.
  (* sigma *)
  + rewrite close_cong_sigma in * by auto.
    rewrite interpV_sigma in *.
    destruct v; simpl in *; try congruence; destruct_things.
    assert (interpV (close_ttyp rho' A1 0) v1) as Hinterp'
      by (apply IHA1 with (rho:=rho); auto).
    intuition.
    rewrite <- open_close_closed_ttyp_swap by (eauto using interpV_closed).
    change (close_ttyp rho' (({0 ~T> v1}) A2) 0)
      with (close_ttyp (v1::rho') A2 0).
    rewrite <- open_close_closed_ttyp_swap in H4 by (eauto using interpV_closed).
    change (close_ttyp rho (({0 ~T> v1}) A2) 0)
      with (close_ttyp (v1::rho) A2 0) in H4.
    apply IHA2 with (rho := v1::rho);  (eauto using interpV_closed, interpV_val). 
Qed.

Lemma interpV_open_joinable : forall (A:ttyp) a a',
  fvs_texp a = 0 -> fvs_texp a' = 0 ->
  joinable a a' ->
    forall v, interpV ({0 ~T> a}  A) v -> interpV ({0 ~T> a'} A) v.
Proof.
  intros A a a' Hfv1 Hfv2 Hj v.
  apply interpV_respects_joinable with (rho := a::nil) (rho' := a'::nil); auto.
Qed.

(* two corollaries of the above *)
Lemma interpV_open_mptstep : forall (A:ttyp) a a',
  fvs_texp a = 0 -> fvs_texp a' = 0 -> 
  mptstep a a' ->
    forall v, (interpV ({0 ~T> a}  A) v <-> interpV ({0 ~T> a'} A) v).
Proof.
  intros.  
  assert (joinable a a').
    exists a'; split; auto.
 intuition;
   eauto using interpV_open_joinable, joinable_symm.
Qed.

Lemma interpV_open_mtstep : forall (A:ttyp) a a',
  fvs_texp a = 0 -> fvs_texp a' = 0 -> 
  mtstep a a' ->
    forall v, (interpV ({0 ~T> a}  A) v <-> interpV ({0 ~T> a'} A) v).
Proof.
  intros.  
  eauto using interpV_open_mptstep, mtstep_mptstep.
Qed.

Lemma exists_iff_thingy : forall (A:Set) (P Q : A -> Prop),
  (forall r, (P r <-> Q r))
  -> ((exists r, P r) <-> (exists r, Q r)).
Proof.
  firstorder.
Qed.

Lemma interpC_open_mtstep : forall (A:ttyp) a a',
  fvs_texp a = 0 -> fvs_texp a' = 0 -> 
  mtstep a a' ->
    forall v, (interpC ({0 ~T> a}  A) v <-> interpC ({0 ~T> a'} A) v).
Proof.
  intros.
  unfold interpC.
   apply exists_iff_thingy.
  intros; split; split; destruct_things; 
    eauto using interpV_open_joinable.
Qed.

Lemma interpV_respects_subtyping :
 forall A A',
   better_ttyp A A' ->
   forall G rho v, 
     wf_env G rho ->
     interpV (close_ttyp rho A 0) v -> interpV (close_ttyp rho A' 0) v.
Proof.
  induction 1; auto; intros G rho v wf_rho.
+ (* Arrow case: *)
    repeat rewrite close_cong_arr by (eauto using wf_env_closed).
    inversion H; subst.
    (* log <: prog : *)
      rewrite interpV_arr_prog.
      rewrite interpV_arr_log.
        destruct v; try tauto; intros [? ?]; constructor; 
          eauto using interpV_val.
    (* th' <: th' : *)
      destruct th'.
       * (* log <: log *)
         repeat rewrite interpV_arr_log.
         (* destruct v; try tauto. *)
         (* texp_lam *)
         unfold interpC; intuition.
         assert (interpV (close_ttyp rho A 0) v0) as H_interpV_A_v0. 
            apply IHbetter_ttyp1 with (G:=G); auto.
         destruct (H4 v0 H_interpV_A_v0) as [r [Hr_interp Hr_step]].
         exists r.  split.
           rewrite <- open_close_closed_ttyp_swap 
             by (eauto using wf_env_closed, interpV_closed).
           change (close_ttyp rho (({0 ~T> v0}) B') 0)
             with (close_ttyp (v0 :: rho) B' 0).
           apply IHbetter_ttyp2 with (G := A' :: G).
             auto.
             simpl.
             rewrite <- open_close_closed_ttyp_swap in Hr_interp
               by (eauto using wf_env_closed, interpV_closed).
             apply Hr_interp.
             auto.            
       * (* prog <: prog *)
         auto.
+ (* Sigma case *)
  intros.
  rewrite close_cong_sigma in * by (eauto using wf_env_closed).
  repeat rewrite interpV_sigma in *.
  destruct v; intuition.    
  rewrite <- open_close_closed_ttyp_swap by (eauto using wf_env_closed, interpV_closed).
  change (close_ttyp rho ({0 ~T> v1} B') 0)
    with (close_ttyp (v1::rho) B' 0).
  apply IHbetter_ttyp with (G := (A::G)).
    auto.
    simpl.    
    rewrite <- open_close_closed_ttyp_swap in H3 by (eauto using wf_env_closed, interpV_closed).
    assumption.
Qed.

Lemma mtstep_interpV_join : forall a v,
  mtstep a v -> interpV (ttyp_eq v a) texp_join.
Proof.
 intros.
 rewrite interpV_eq.
 split; auto.
Qed.

Lemma wf_env_length : forall G rho,
  wf_env G rho ->
  length G = length rho.
Proof.
  induction 1; simpl; auto.
Qed.

Lemma le_pred' : forall n m,
  n <= S m -> pred n <= m.
Proof.
  intros. change m with (pred (S m)).
  apply le_pred; assumption.
Qed.

Lemma fvs_in_context :
 forall G th a A,
   ttyping G th a A ->
     fvs_texp a <= length G.
Proof.
  induction 1; simpl; try rewrite app_length; simpl in *; 
   (try apply max_lub); repeat apply le_pred'; auto; try omega; auto.

   apply max_lub; repeat apply le_pred'; auto.
Qed. 

Lemma fvs_open : 
  forall a n m,
    n <= m ->
    fvs_texp a <= S m ->
     forall b,
       fvs_texp b = 0 -> 
         fvs_texp ({n ~t> b}a) <= m.
Proof.
  induction a; intros; simpl in *;
    repeat apply le_pred'; 
      try apply IHa;  
      try rewrite lift_closed_texp_id by auto;
      repeat  match goal with
           [ H: max _ _ <= _ |- _ ] =>
              let H1 := fresh H in 
              let H2 := fresh H in 
              assert (H1 := max_lub_l _ _ _ H);
              assert (H2 := max_lub_r _ _ _ H);
              clear H
          end;
      try apply max_lub; 
      try apply IHa1;  try apply IHa2;
      try rewrite lift_closed_texp_id by auto;
      auto.

    db_var_case.

    apply max_lub.
      apply le_pred'; apply IHa2; auto.
      repeat apply le_pred'; apply IHa3; auto.

    repeat apply le_pred'; apply IHa2; auto.
Qed.

Lemma close_texp_closes_gen' : 
 forall rho,
  closed_env rho -> 
   forall a n, 
     fvs_texp a <= length rho + n  ->
     fvs_texp (close_texp rho a n) <= n.
Proof.
 induction 1; simpl; intros; auto.

  apply IHclosed_env.
  apply fvs_open; auto.
Qed.

Lemma close_texp_closes_gen : 
  forall G th a A ,
   ttyping G th a A ->  
    forall rho n,
     closed_env rho ->
     n + length rho = length G ->
     fvs_texp (close_texp rho a n) <= n.
Proof.
  intros.
  apply fvs_in_context in H.
  rewrite <- H1 in H; clear H1.
  apply close_texp_closes_gen'; auto.
Qed.

Lemma close_texp_closes : 
  forall G th a A rho,
     ttyping G th a A -> 
     wf_env G rho
     -> fvs_texp (close_texp rho a 0) = 0.
Proof.
 intros.
 symmetry; apply le_n_0_eq.
 apply close_texp_closes_gen with (G:=G) (th:=th) (A:=A).
   auto.
   eauto using wf_env_closed.
   rewrite (wf_env_length G rho); auto.
Qed.

Lemma close_texp_closes_lam : 
  forall A1 G th a A rho,
     ttyping (A1 :: G) th a A -> 
     wf_env G rho
     -> fvs_texp (close_texp rho (texp_lam a) 0) = 0.
Proof.
 intros.
 symmetry; apply le_n_0_eq.
 rewrite close_cong_lam by (eauto using wf_env_closed).
 simpl.
 change 0 with (pred 1); apply le_pred; simpl.
 apply close_texp_closes_gen with (G:=A1::G) (th:=th) (A:=A).
   auto.
   eauto using wf_env_closed.
   simpl; rewrite (wf_env_length G rho); auto.
Qed.

Lemma close_texp_closes_fun : 
  forall A1 A2 G th a A rho,
     ttyping (A1 :: A2 :: G) th a A -> 
     wf_env G rho
     -> fvs_texp (close_texp rho (texp_fun a) 0) = 0.
Proof.
 intros.
 symmetry; apply le_n_0_eq.
 rewrite close_cong_fun by (eauto using wf_env_closed).
 simpl.
 change 0 with (pred (pred 2)); apply le_pred; apply le_pred; simpl.
 apply close_texp_closes_gen with (G:=A1::A2::G) (th:=th) (A:=A).
   auto.
   eauto using wf_env_closed.
   simpl; rewrite (wf_env_length G rho); auto.
Qed.

Lemma close_texp_closes_nrec : 
  forall A1 A2 G th a A rho,
     ttyping (A1 :: A2 :: G) th a A -> 
     wf_env G rho
     -> fvs_texp (close_texp rho (texp_nrec a) 0) = 0.
Proof.
 intros.
 symmetry; apply le_n_0_eq.
 rewrite close_cong_nrec by (eauto using wf_env_closed).
 simpl.
 change 0 with (pred (pred 2)); apply le_pred; apply le_pred; simpl.
 apply close_texp_closes_gen with (G:=A1::A2::G) (th:=th) (A:=A).
   auto.
   eauto using wf_env_closed.
   simpl; rewrite (wf_env_length G rho); auto.
Qed.


Lemma soundness_var : 
 forall (G1 G3:tctx) (x:nat) (A:ttyp),
  x = length G1 ->
  forall rho (Hwf_rho : wf_env (G1 ++ A :: G3) rho),
   interpV (close_ttyp rho (lift_ttyp A (S x) 0) 0) (close_texp rho (texp_var x) 0). 
Proof.
  induction G1.
  + (* G1 := nil *) intros.
     simpl in *.
     inversion Hwf_rho.
     subst. simpl in *.
     replace (close_texp rho' a 0) with a
       by (rewrite close_closed_texp_id; eauto using interpV_closed).
     autorewrite with step.
     assumption.

  + (* G1 := a::G1 *)
    intros.
    simpl in *.
    inversion Hwf_rho.
    simpl.
    case_eq (nat_compare x 0); intros x_eq.
      - apply nat_compare_eq in x_eq.
        omega.
      - apply nat_compare_lt in x_eq.
        omega.
      - apply nat_compare_gt in x_eq.
        destruct x as [|y] ; [ omega |idtac].
        autorewrite with step.
        apply IHG1 with (G3:=G3).
          omega.
          assumption.
Qed.

Ltac inst_soundness_IH :=
match goal with
 [ H1 : ?th1 = th_log -> forall (rho':env), wf_env ?G rho' -> interpC _ _,                                            
(*         exists (r : texp), _ /\ _ , *)
   H2 : wf_env ?G ?rho
   |- _ ] => 
     let v := fresh "v" in
     let m := fresh "m" in
     let Hth_is_log := fresh "Hth_is_log" in
     let Hinterp := fresh "Hinterp" in
     let Hsteps := fresh "Hsteps" in
       assert (th1 = th_log) as Hth_is_log by congruence;
       destruct (H1 Hth_is_log rho H2) as [v [Hinterp Hsteps]];
       clear H1;
       clear Hth_is_log
end.


Lemma interpV_interpC : forall A v,
  interpV A v -> interpC A v.
Proof.
  intros.
  exists v.
  eauto.
Qed.

Lemma interpC_mtstep : forall A a a',
  interpC A a' -> mtstep a a'  -> interpC A a.
Proof.
  intros A a a' [v [H1 H2]] Hsteps.
  exists v. split. 
    assumption.
    apply MSTrans with a'; auto.
Qed.      

Lemma interpC_tstep : forall A a a',
  interpC A a' -> tstep a a' -> interpC A a.
Proof.
  intros. eauto using interpC_mtstep, tstep_mtstep.
Qed.      

Lemma interpV_lam : forall A1 A2 b,
  fvs_texp b <= 1 ->
  (forall x, interpV A1 x -> interpC ({0 ~T> x}A2) ({0 ~t> x}b)) ->
  interpV (ttyp_arr th_log A1 A2) (texp_lam b).
Proof.
  intros.
  rewrite interpV_arr_log.
  split.
    + split; simpl; auto; omega.
    + intros v Hinterp.
      apply interpC_tstep with (({0 ~t> v}) b).
        auto.
        eauto using interpV_val.
Qed.

Lemma joinable_numerals : 
  forall n,  tnumeral n 
   -> forall m, tnumeral m
        -> joinable n m -> n = m.
Proof.
  induction 1.
  + inversion 1; subst.
    - auto.
    - intros [c [H1 H2]].
      apply mptstep_zero_inversion in H1.
      apply mptstep_succ_inversion in H2.
      destruct_things; congruence.
  + inversion 1; subst.
    - intros [c [H1 H2]].
      apply mptstep_succ_inversion in H1.
      apply mptstep_zero_inversion in H2.
      destruct_things; congruence.
    - intros [c [Ha Ha0]].
      apply mptstep_succ_inversion in Ha.
      apply mptstep_succ_inversion in Ha0.
      destruct_things.
      assert (x = x0) by congruence; subst.
      f_equal.
      apply IHtnumeral; auto.
        exists x0; auto.
Qed.

Lemma interpV_eq_numerals : forall n m v,
  tnumeral n -> tnumeral m  ->
    interpV (ttyp_eq n m) v
    -> n = m.
Proof.
  intros n m v Hn Hm Hinterp.
  rewrite interpV_eq in Hinterp.  
  destruct Hinterp as [_ Hjoinable].
  auto using joinable_numerals.
Qed.

Theorem soundness : forall G th_top a A,
  ttyping G th_top a A ->
  th_top = th_log ->
  forall (rho: env) (Hwf_rho : wf_env G rho),
    interpC (close_ttyp rho A 0) (close_texp rho a 0).
Proof.

Hint Constructors mtstep.
Hint Constructors wf_env.

induction 1.

(* TTVar *)
+ intros.
  exists (close_texp rho (texp_var n) 0).
  eauto using soundness_var. 

(* TTApp *)
+ intros.
  repeat inst_soundness_IH.
  rewrite close_cong_arr in Hinterp0 by (eauto using wf_env_closed).
  destruct th2.
   * apply interpV_arr_log in Hinterp0.
     destruct Hinterp0 as [Hv0_closed Hinterp0].
     rewrite open_close_ttyp_swap by (eauto using wf_env_closed).
     rewrite close_cong_app.
     apply interpC_mtstep with (texp_app v0 (close_texp rho b 0)).
       apply interpC_mtstep with (texp_app v0 v).
         rewrite interpC_open_mtstep with (a' := v)
          by (eauto using close_texp_closes, interpV_closed, interpV_val).
         apply Hinterp0; auto.

         apply MSApp2; auto.
       apply MSApp1; assumption.
  * match goal with
     [ H: better_theta th_prog _ |- _ ] => inversion H; congruence
    end.

(* TTLam *)
+ intros.
  apply interpV_interpC.
  rewrite close_cong_arr by (eauto using wf_env_closed).
  destruct th2.
  * (* logical lambda *)
   rewrite interpV_arr_log.
   split.
   - split.
       rewrite close_cong_lam  by (eauto using wf_env_closed). auto.
       eauto using close_texp_closes_lam.
   - rewrite close_cong_lam by (eauto using wf_env_closed); simpl.
     intros v Hinterp_v.
     assert (wf_env (A::G) (v::rho)) by auto.
     inst_soundness_IH.
     simpl in Hsteps.
     rewrite open_close_closed_texp_swap  in Hsteps
       by (eauto using wf_env_closed, interpV_closed). 

     apply interpC_tstep with (({0 ~t> v}) (close_texp rho b 1)).       
      { exists v0.
         split.
           rewrite <- open_close_closed_ttyp_swap 
              by (eauto using wf_env_closed, interpV_closed). 
           exact Hinterp.

           assumption. }
      { apply TSBetaLam.
          eauto using interpV_val. }
  * (* Programmatic lambda *)
    rewrite interpV_arr_prog.
    split. 
      rewrite close_cong_lam by (eauto using wf_env_closed); auto.
      eauto using close_texp_closes_lam.

(* TTRec *)
+ intros.
  apply interpV_interpC.
  rewrite close_cong_arr by (eauto using wf_env_closed).
  rewrite interpV_arr_prog.
  split. 
    rewrite close_cong_fun by (eauto using wf_env_closed); auto.
    eauto using close_texp_closes_fun.
      
(* TTNRec *)
+ intros.
  apply interpV_interpC.
  rewrite close_cong_arr, interpV_arr_log by (eauto using wf_env_closed).
  split. 
    { split. 
        rewrite close_cong_nrec by (eauto using wf_env_closed); auto.
        eauto using close_texp_closes_nrec. }
  rewrite close_cong_nrec by (eauto using wf_env_closed).
  intros v Hnumeral.
  rewrite close_cong_nat, interpV_nat in Hnumeral.
  induction Hnumeral.
  * (* zero *)


   assert (wf_env  (ttyp_arr th_log ttyp_nat
                     (ttyp_arr th_log
                        (ttyp_eq (texp_succ (texp_var 0)) (texp_var 1))
                        (lift_ttyp (lift_ttyp B 1 1) 1 0))
                     :: ttyp_nat :: G) 
                   ((texp_lam
                      (texp_lam
                         (texp_app (texp_nrec (lift_texp (close_texp rho a 2) 3 2))
                                   (texp_var 1))))
                      :: texp_zero :: rho)).
   apply ECons.
     apply ECons; auto.
     rewrite close_cong_nat, interpV_nat; auto.
     { 
       repeat rewrite close_cong_arr  by (eauto 7 using wf_env_closed).
       rewrite close_cong_nat.
       rewrite close_cong_eq.
       rewrite close_cong_succ.
       simpl.
       rewrite close_cong_var_gt by omega.
       apply interpV_lam.
         {
           simpl. 
           apply le_pred'.
           apply Max.max_lub.
           + repeat apply le_pred'. 
             rewrite lift_closed_texp_id_gen
             by (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
             apply le_trans with 2; try omega; 
               (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
           + omega.
         }
         
         intros n Hnumeral_n.
         rewrite interpV_nat in Hnumeral_n.
         apply interpV_interpC.
         simpl.
         apply interpV_lam.
         {
           simpl. apply Max.max_lub.
           + repeat apply le_pred'. 
             rewrite open_texp_under_lift by omega.
             rewrite lift_closed_texp_id_gen
             by (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
             apply le_trans with 2; try omega; 
               (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
           + assert (fvs_texp n = 0) by eauto using interpV_closed.
             rewrite lift_closed_texp_id;
               [ apply le_trans with 0; try omega | idtac]; auto.
         }
         { intros join Hjoin.
           rewrite close_closed_texp_id in Hjoin by (simpl;auto).
           rewrite open_closed_texp_id in Hjoin by (simpl;auto).
           apply interpV_eq_numerals in Hjoin; auto.
           congruence.
         }
     }

  eapply interpC_tstep; [idtac | apply TSBetaNRec; auto].
  inst_soundness_IH.
  simpl in Hinterp.
  rewrite open_ttyp_under_lift in Hinterp by omega.
  rewrite lift_ttyp_0_is_id in Hinterp.
  rewrite open_close_ttyp_swap in Hinterp by (eauto using wf_env_closed).
  rewrite close_closed_texp_id in Hinterp by (eauto using interpV_closed).
  exists v. split.
    - assumption.
    - simpl in Hsteps.
      repeat rewrite open_close_texp_swap in Hsteps by (eauto using wf_env_closed).
      repeat rewrite close_cong_lam in Hsteps  by (eauto using wf_env_closed).
      rewrite close_cong_app in Hsteps.
      rewrite close_cong_nrec in Hsteps  by (eauto using wf_env_closed).
      rewrite close_cong_var_gt in Hsteps by omega.
      rewrite lift_closed_texp_id_gen in Hsteps
             by (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
      rewrite lift_closed_texp_id_gen 
             by (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]). 
      replace (close_texp rho texp_zero 0) with texp_zero in Hsteps
        by (rewrite close_closed_texp_id; simpl; auto).
      replace  (close_texp rho (close_texp rho a 2) 5) 
        with (close_texp rho a 2) in Hsteps
        by (symmetry;
            apply close_closed_texp_id_gen;
            apply le_trans with 2; try omega;
            eapply close_texp_closes_gen with (G:=(ttyp_arr th_log ttyp_nat
              (ttyp_arr th_log (ttyp_eq (texp_succ (texp_var 0)) (texp_var 1))
                 (lift_ttyp (lift_ttyp B 1 1) 1 0)) :: 
            ttyp_nat :: G)); eauto using wf_env_closed; simpl;
              rewrite (wf_env_length _ _ Hwf_rho); auto).
      assumption.

  * (* succ *)
   assert (wf_env  (ttyp_arr th_log ttyp_nat
                     (ttyp_arr th_log
                        (ttyp_eq (texp_succ (texp_var 0)) (texp_var 1))
                        (lift_ttyp (lift_ttyp B 1 1) 1 0))
                     :: ttyp_nat :: G) 
                   ((texp_lam
                      (texp_lam
                         (texp_app (texp_nrec (lift_texp (close_texp rho a 2) 3 2))
                                   (texp_var 1))))
                      :: texp_succ a0 :: rho)).
   apply ECons.
     apply ECons; auto.
     rewrite close_cong_nat, interpV_nat; auto.
     (* "And now for the tricky bit." *)
     { 
       repeat rewrite close_cong_arr  by (eauto 7 using wf_env_closed).
       rewrite close_cong_nat.
       rewrite close_cong_eq.
       rewrite close_cong_succ.
       simpl.
       rewrite close_cong_var_gt by omega.
       apply interpV_lam.
         {
           simpl. 
           apply le_pred'.
           apply Max.max_lub.
           + repeat apply le_pred'. 
             rewrite lift_closed_texp_id_gen
             by (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
             apply le_trans with 2; try omega; 
               (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
           + omega.
         }
         
         intros n Hnumeral_n.
         rewrite interpV_nat in Hnumeral_n.
         apply interpV_interpC.
         simpl.
         apply interpV_lam.
         {
           simpl. apply Max.max_lub.
           + repeat apply le_pred'. 
             rewrite open_texp_under_lift by omega.
             rewrite lift_closed_texp_id_gen
             by (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
             apply le_trans with 2; try omega; 
               (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
           + assert (fvs_texp n = 0) by eauto using interpV_closed.
             rewrite lift_closed_texp_id;
               [ apply le_trans with 0; try omega | idtac]; auto.
         }
         { intros join Hjoin.
           simpl.
           (* I do not have words to express how much I hate de Bruijn indices: *)
           repeat rewrite open_texp_under_lift by omega.
           repeat rewrite lift_texp_0_is_id.
           rewrite lift_lift_ttyp_swap_0.
           rewrite open_ttyp_under_lift by omega.
           rewrite lift_ttyp_0_is_id.
           rewrite close_lift_ttyp_swap by (eauto using wf_env_closed).
           rewrite open_lift_ttyp_commute_1 by omega.
           rewrite open_ttyp_under_lift by omega.
           rewrite lift_texp_0_is_id.
           rewrite lift_ttyp_0_is_id.
           (* Phew! *)

           rewrite close_closed_texp_id in Hjoin by (simpl;auto).
           rewrite open_closed_texp_id in Hjoin by (simpl;auto).
           apply interpV_eq_numerals in Hjoin; auto.
           replace n with a0 by congruence.

           rewrite lift_closed_texp_id_gen
             by (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
           exact IHHnumeral.
         }
      }
  clear IHHnumeral.
  eapply interpC_tstep; [idtac | apply TSBetaNRec; auto].
  inst_soundness_IH.
  simpl in Hinterp.
  rewrite open_ttyp_under_lift in Hinterp by omega.
  rewrite lift_ttyp_0_is_id in Hinterp.
  rewrite open_close_ttyp_swap in Hinterp by (eauto using wf_env_closed).
  rewrite close_closed_texp_id in Hinterp by (eauto using interpV_closed).
  exists v. split.
    - assumption.
    - simpl in Hsteps.
      repeat rewrite open_close_texp_swap in Hsteps by (eauto using wf_env_closed).
      repeat rewrite close_cong_lam in Hsteps  by (eauto using wf_env_closed).
      rewrite close_cong_app in Hsteps.
      rewrite close_cong_nrec in Hsteps  by (eauto using wf_env_closed).
      rewrite close_cong_var_gt in Hsteps by omega.
      rewrite lift_closed_texp_id_gen in Hsteps
             by (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]).
      rewrite lift_closed_texp_id_gen 
             by (apply (close_texp_closes_gen _ _ _ _ H);
                  [ eauto using wf_env_closed
                  | simpl; rewrite (wf_env_length _ _ Hwf_rho); auto]). 
      replace (close_texp rho (texp_succ a0) 0) with (texp_succ a0) in Hsteps
        by (rewrite close_closed_texp_id; simpl; auto).
      replace  (close_texp rho (close_texp rho a 2) 5) 
        with (close_texp rho a 2) in Hsteps
        by (symmetry;
            apply close_closed_texp_id_gen;
            apply le_trans with 2; try omega;
            eapply close_texp_closes_gen with (G:=(ttyp_arr th_log ttyp_nat
              (ttyp_arr th_log (ttyp_eq (texp_succ (texp_var 0)) (texp_var 1))
                 (lift_ttyp (lift_ttyp B 1 1) 1 0)) :: 
            ttyp_nat :: G)); eauto using wf_env_closed; simpl;
              rewrite (wf_env_length _ _ Hwf_rho); auto).
      assumption.

(* TTZero *)
+ intros.
  apply interpV_interpC.
  rewrite close_cong_nat.
  rewrite interpV_nat.
  rewrite close_cong_zero.
  auto.

(* TTSuc *)
+ intros.
  inst_soundness_IH.
  exists (texp_succ v).
  split.
  * rewrite close_cong_nat in *.
    rewrite interpV_nat in *. 
    auto.
  * rewrite close_cong_succ.
    apply MSSucc. auto.

(* TTPair *)
+ intros.
  repeat inst_soundness_IH.
  exists (texp_pair v0 v).
  split.  
  * rewrite close_cong_sigma by (eauto using wf_env_closed).
    rewrite interpV_sigma.
    split.
      unfold closed_tval. simpl. split.
        eauto using interpV_val.
        assert (fvs_texp v0 = 0) by (eauto using interpV_closed).
        assert (fvs_texp v = 0) by (eauto using interpV_closed).
        symmetry; apply le_n_0_eq; apply Max.max_lub; omega.
      split.
      - auto.        
      - rewrite open_close_ttyp_swap in Hinterp
              by (eauto using wf_env_closed, interpV_closed).
        rewrite interpV_open_mtstep with (a' := v0) in Hinterp.
          assumption.
          eauto using close_texp_closes.
          eauto using interpV_closed.
          eauto using interpV_val.

  * rewrite close_cong_pair.
    apply MSPair.
      assumption.
      assumption.
      eauto using interpV_val.

(* TTPCase *)
+ intros.
  inst_soundness_IH.
  rewrite close_cong_sigma in Hinterp by (eauto using wf_env_closed).
  rewrite interpV_sigma in Hinterp.
  destruct v as [ ? | ? | ? | ? | ? |  | ? | ? | v1 v2 | ? | ];
    simpl in Hinterp; try tauto.
  destruct_things.
  assert (wf_env  (ttyp_eq (texp_pair (texp_var 1) (texp_var 0))
                    (lift_texp a 2 0) :: A2 :: A1 :: G)
                  (texp_join :: v2 :: v1 :: rho)).
    apply ECons.
      apply ECons.
        apply ECons.
          auto.
          auto.
        simpl;
          rewrite open_close_closed_ttyp_swap 
              by (eauto using wf_env_closed, interpV_closed);
          auto.
      simpl; autorewrite with step.
      rewrite open_closed_texp_id
         by (eauto using interpV_closed).
      repeat rewrite close_closed_texp_id
         by (eauto using wf_env_closed, interpV_closed, interpV_val).
      eapply mtstep_interpV_join; eassumption.

  inst_soundness_IH.
  exists v.
  split.
  * simpl in Hinterp.
    repeat rewrite open_lift_ttyp_id in Hinterp.
    rewrite lift_ttyp_0_is_id in Hinterp.
    assumption.
  * simpl in Hsteps0.
    apply MSTrans with (close_texp rho
                 (({0 ~t> v1}) (({0 ~t> v2}) (({0 ~t> texp_join}) b))) 0); auto.
    apply MSTrans with (texp_pcase (texp_pair v1 v2) (close_texp rho b 3)).

    rewrite close_cong_pcase by (eauto using wf_env_closed).
    apply MSPCase1; assumption.

    apply tstep_mtstep.  
    repeat rewrite open_close_closed_texp_swap in *
       by (eauto using interpV_closed, wf_env_closed).
    pattern v2 at 2. 
    replace v2 with (lift_texp v2 1 0) 
      by (eauto using lift_closed_texp_id, interpV_closed).
    apply TSCasePair;
      eauto using (interpV_val).

(* TTNCase *)
+ intros.
  inst_soundness_IH.
  rewrite close_cong_nat, interpV_nat in Hinterp.
  destruct Hinterp.
  - (* texp_zero *)
    clear IHttyping3.
    assert (wf_env  (ttyp_eq texp_zero a :: G)
                    (texp_join :: rho)). 
      apply ECons.
        auto.
        rewrite close_cong_eq, interpV_eq.
        split; auto.
        rewrite close_cong_zero.
        apply mtstep_interpV_join; auto.
   inst_soundness_IH.
   exists v.
   split.
   * simpl in Hinterp.
     repeat rewrite open_lift_ttyp_id in Hinterp.
     rewrite lift_ttyp_0_is_id in Hinterp.
     assumption.
   * rewrite close_cong_ncase by (eauto using wf_env_closed).
     simpl in *.
     apply MSTrans with (close_texp rho (({0 ~t> texp_join}) b1) 0); auto.
     apply MSTrans with (texp_ncase texp_zero 
                                    (close_texp rho b1 1) (close_texp rho b2 2)). 
     apply MSNatCase1; auto.
     
     rewrite open_close_closed_texp_swap by (eauto using wf_env_closed).
     apply tstep_mtstep;  apply TSCaseZ.      
  - (* texp_succ *)
    clear IHttyping2.
    assert (wf_env  (ttyp_eq (texp_succ (texp_var 0)) (lift_texp a 1 0)
                      :: ttyp_nat :: G)
                    (texp_join :: a0 :: rho)). 
      { apply ECons.
          auto.
          apply ECons.
            auto.
            rewrite close_cong_nat, interpV_nat; auto.

          rewrite close_cong_eq.
          apply mtstep_interpV_join.
          simpl. rewrite open_lift_texp_id. rewrite lift_texp_0_is_id.
          rewrite close_closed_texp_id with (a := texp_succ a0).
            auto.
            eauto using interpV_closed. }
    inst_soundness_IH.
    exists v.
    split.
    * simpl in Hinterp0.
      repeat rewrite open_lift_ttyp_id in Hinterp0.
      rewrite lift_ttyp_0_is_id in Hinterp0.
      assumption.
    * rewrite close_cong_ncase by (eauto using wf_env_closed).
      simpl in *.
      apply MSTrans 
        with  (close_texp rho (({0 ~t> a0}) (({0 ~t> texp_join}) b2)) 0); auto.
      apply MSTrans with (texp_ncase (texp_succ a0)
                                     (close_texp rho b1 1) (close_texp rho b2 2)). 
      apply MSNatCase1; auto.

      repeat rewrite open_close_closed_texp_swap by (eauto using wf_env_closed).
      apply tstep_mtstep;  apply TSCaseS; eauto using interpV_val.

(* TTJoin *)
+ intros.
  apply interpV_interpC.
  rewrite close_cong_eq.
  rewrite interpV_eq.
  rewrite close_closed_texp_id; auto.
  split; auto.
  exists (close_texp rho c 0); 
    eauto using close_preserves_mptstep.

(* TTConv *)
+ intros.
  clear IHttyping2.
  clear IHttyping3.
  repeat inst_soundness_IH.
  exists v.
  split.
  * rewrite close_cong_eq in Hinterp0.
    rewrite interpV_eq in Hinterp0.
    destruct_things.
 
    rewrite open_close_ttyp_swap by (eauto using wf_env_closed).
    apply interpV_open_joinable with (close_texp rho b1 0).
      eauto using interpV_closed.
      eauto using close_texp_closes.
      eauto using close_texp_closes.
      assumption.     
      rewrite <- open_close_ttyp_swap by (eauto using wf_env_closed).
      assumption.
  * assumption. 

(* TTSubtype *)
+ intros.
  inst_soundness_IH.
  exists v.
  split.
  * apply interpV_respects_subtyping with (G:=G) (A:=A); auto.
  * auto.
Qed.

Corollary normalization : forall a A,
  ttyping nil th_log a A ->
    exists v, tval v /\ mtstep a v.
Proof.
  intros a A H.
  apply soundness with (rho := nil) in H ; auto.
  simpl in H.
  destruct H as [v [H1 H2]].
  exists v.
    eauto using interpV_val.
Qed.

Corollary consistency : forall a,
  ~ ttyping nil th_log a (ttyp_eq texp_zero (texp_succ texp_zero)).
Proof.
  intros a H.
  apply soundness with (rho := nil) in H ; auto.
  simpl in H.
  destruct H as [v [H _]].
  apply interpV_eq_numerals in H; auto; congruence.
Qed.

Print Assumptions soundness.